# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-08 18:33
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0002_auto_20160907_1912'),
    ]

    operations = [
        migrations.AddField(
            model_name='unshardedtestmodel',
            name='more_random_string',
            field=models.CharField(default=11, max_length=120),
            preserve_default=False,
        ),
    ]



from random import choice
from uuid import UUID

from django.conf import settings
from django.test import TestCase
from django.utils.six.moves import xrange

from tests.models import ShardedModelIDs
from django_sharding_library.id_generation_strategies import TableStrategy, UUIDStrategy


class TableStrategyIDGenerationTestCase(TestCase):

    def test_returns_unique_values(self):
        sut = TableStrategy(ShardedModelIDs)
        ids = [sut.get_next_id() for i in xrange(100)]
        self.assertEqual(ids, list(set(ids)))

    def test_largest_value_stored_in_db(self):
        sut = TableStrategy(ShardedModelIDs)
        for i in xrange(100):
            id = sut.get_next_id()
            self.assertEqual(ShardedModelIDs.objects.latest('pk').pk, id)
            self.assertFalse(ShardedModelIDs.objects.filter(pk__gt=id).exists())


class UUIDStrategyTestCase(TestCase):

    def test_uuid_strategy_must_be_passed_a_database(self):
        sut = UUIDStrategy()
        with self.assertRaises(AssertionError):
            sut.get_next_id('im not a database')

    def test_returns_value_with_db_name_and_uuid(self):
        sut = UUIDStrategy()
        for i in xrange(100):
            database = choice(list(settings.DATABASES.keys()))
            id = sut.get_next_id(database)
            self.assertTrue(id.startswith(database))
            uuid_value = id[len(database) + 1:]
            self.assertEqual(str(UUID(uuid_value, version=4)), uuid_value)



from django.db import models
from django_sharding_library.id_generation_strategies import TableStrategyModel
from django_sharding_library.decorators import model_config
from django_sharding_library.exceptions import NonExistentDatabaseException, ShardedModelInitializationException
from django_sharding_library.fields import TableShardedIDField
from django.test import TestCase
from django_sharding_library.manager import ShardManager


class ModelConfigDecoratorTestCase(TestCase):

    def setUp(self):
        class ShardedDecoratorTestModelIDs(TableStrategyModel):
            pass
        self.id_table = ShardedDecoratorTestModelIDs

    def test_model_cannot_be_both_sharded_and_marked_for_a_specific_db(self):
        with self.assertRaises(ShardedModelInitializationException):
            @model_config(shard_group='default', database='default')
            class TestModelTwo(models.Model):
                id = TableShardedIDField(primary_key=True, source_table=self.id_table)
                random_string = models.CharField(max_length=120)
                user_pk = models.PositiveIntegerField()

                def get_shard(self):
                    pass

    def test_sharded_model_requires_a_get_shard_method(self):
        with self.assertRaises(ShardedModelInitializationException):
            @model_config(shard_group='default')
            class TestModelTwo(models.Model):
                id = TableShardedIDField(primary_key=True, source_table=self.id_table)
                random_string = models.CharField(max_length=120)
                user_pk = models.PositiveIntegerField()

    def test_sharded_id_field_must_be_primary_key(self):
        with self.assertRaises(ShardedModelInitializationException):
            @model_config(shard_group='default')
            class TestModelTwo(models.Model):
                id = TableShardedIDField(source_table=self.id_table)
                random_string = models.CharField(max_length=120)
                user_pk = models.PositiveIntegerField(primary_key=True)

                def get_shard(self):
                    pass

    def test_sharded_model_must_have_sharded_id_fied(self):
        with self.assertRaises(ShardedModelInitializationException):
            @model_config(shard_group='default')
            class TestModelTwo(models.Model):
                random_string = models.CharField(max_length=120)
                user_pk = models.PositiveIntegerField()

                def get_shard(self):
                    from django.contrib.auth import get_user_model
                    return get_user_model().objects.get(pk=self.user_pk).shard

    def test_puts_shard_group_on_the_model_class(self):
        @model_config(shard_group='testing')
        class TestModelThree(models.Model):
            id = TableShardedIDField(source_table=self.id_table, primary_key=True)
            random_string = models.CharField(max_length=120)
            user_pk = models.PositiveIntegerField()

            def get_shard(self):
                from django.contrib.auth import get_user_model
                return get_user_model().objects.get(pk=self.user_pk).shard

        self.assertEqual(getattr(TestModelThree, 'django_sharding__shard_group', None), 'testing')

    def test_cannot_place_database_on_replica_db(self):
        with self.assertRaises(NonExistentDatabaseException):
            @model_config(database='app_shard_001_replica_001')
            class ShardedTestModelIDsTwo(TableStrategyModel):
                pass

    def test_cannot_place_database_on_non_existant_db(self):
        with self.assertRaises(NonExistentDatabaseException):
            @model_config(database='i_do_not_exist')
            class ShardedTestModelIDsTwo(TableStrategyModel):
                pass

    def test_puts_database_name_on_model_stored_on_another_database(self):
        @model_config(database='app_shard_002')
        class ShardedTestModelIDsThree(TableStrategyModel):
            pass

        self.assertEqual(getattr(ShardedTestModelIDsThree, 'django_sharding__database', None), 'app_shard_002')

    def test_abstract_model_with_defined_manager_raises_exception_if_not_instance_of_shard_manager(self):
        # Manager is defined and not shard model, should raise an exception
        with self.assertRaises(ShardedModelInitializationException):
            @model_config(shard_group='default', sharded_by_field="user_pk")
            class TestModelOne(models.Model):
                id = TableShardedIDField(primary_key=True, source_table=self.id_table)
                random_string = models.CharField(max_length=120)
                user_pk = models.PositiveIntegerField()

                objects = models.Manager()

                class Meta:
                    abstract = True

                def get_shard(self):
                    pass

                @staticmethod
                def get_shard_from_id(id):
                    pass

        # Manager is not defined, should NOT raise an exception
        @model_config(shard_group='default', sharded_by_field="user_pk")
        class TestModelTwo(models.Model):
            id = TableShardedIDField(primary_key=True, source_table=self.id_table)
            random_string = models.CharField(max_length=120)
            user_pk = models.PositiveIntegerField()

            class Meta:
                abstract = True

            def get_shard(self):
                pass

            @staticmethod
            def get_shard_from_id(id):
                pass

        # Manager is defines but is a shardmanager, should not raise an exception
        @model_config(shard_group='default', sharded_by_field="user_pk")
        class TestModelThree(models.Model):
            id = TableShardedIDField(primary_key=True, source_table=self.id_table)
            random_string = models.CharField(max_length=120)
            user_pk = models.PositiveIntegerField()

            objects = ShardManager()

            class Meta:
                abstract = True

            def get_shard(self):
                pass

            @staticmethod
            def get_shard_from_id(id):
                pass

    def test_decorator_raises_exception_when_sharded_by_field_is_defined_with_no_get_shard_from_id_function(self):
        with self.assertRaises(ShardedModelInitializationException):
            @model_config(shard_group='default', sharded_by_field="user_pk")
            class TestModelOne(models.Model):
                id = TableShardedIDField(primary_key=True, source_table=self.id_table)
                random_string = models.CharField(max_length=120)
                user_pk = models.PositiveIntegerField()

                def get_shard(self):
                    pass

    def test_decorator_raises_exception_when_using_sharded_by_field_and_custom_manager_is_not_shard_manager_instance(self):
        class CustomManager(models.Manager):
            pass

        with self.assertRaises(ShardedModelInitializationException):
            @model_config(shard_group='default', sharded_by_field="user_pk")
            class TestModelOne(models.Model):
                id = TableShardedIDField(primary_key=True, source_table=self.id_table)
                random_string = models.CharField(max_length=120)
                user_pk = models.PositiveIntegerField()

                objects = CustomManager()

                def get_shard(self):
                    pass

                @staticmethod
                def get_shard_from_id(id):
                    pass

    def test_decorator_raises_exception_when_no_arguments_passed_in(self):
        with self.assertRaises(ShardedModelInitializationException):
            @model_config()
            class TestModelOne(models.Model):
                id = TableShardedIDField(primary_key=True, source_table=self.id_table)
                random_string = models.CharField(max_length=120)
                user_pk = models.PositiveIntegerField()

                def get_shard(self):
                    pass

                @staticmethod
                def get_shard_from_id(id):
                    pass



from django.test import TestCase


class TestSaveShardHandler(TestCase):

    def test_shard_is_set(self):
        from django.contrib.auth import get_user_model
        User = get_user_model()
        user = User.objects.create(username='test', password='test')
        self.assertIsNotNone(user.shard)



import django
from django.conf import settings
from django.apps import apps
from django_sharding_library.constants import Backends
from django.utils.six import iteritems
from django.db.models import Manager

from django_sharding_library.exceptions import NonExistentDatabaseException, ShardedModelInitializationException
from django_sharding_library.manager import ShardManager
from django_sharding_library.fields import ShardedIDFieldMixin, PostgresShardGeneratedIDField
from django_sharding_library.utils import register_migration_signal_for_model_receiver

PRE_MIGRATION_DISPATCH_UID = "PRE_MIGRATE_FOR_MODEL_%s"


def model_config(shard_group=None, database=None, sharded_by_field=None):
    """
    A decorator for marking a model as being either sharded or stored on a
    particular database. When sharding, it does some verification to ensure
    that the model is defined correctly.
    """
    def configure(cls):
        if database and shard_group:
            raise ShardedModelInitializationException('A model cannot be both sharded and stored on a particular database.')

        if not database and not shard_group:
            raise ShardedModelInitializationException('The model should be either sharded or stored on a database in the `model_config` decorator is used.')

        if database:
            if database not in settings.DATABASES or settings.DATABASES[database].get('PRIMARY'):
                raise NonExistentDatabaseException(
                    'Unable to place {} in {} as that is not an existing primary database in the system.'.format(cls._meta.model_name, database)
                )
            setattr(cls, 'django_sharding__database', database)

        postgres_shard_id_fields = list(filter(lambda field: issubclass(type(field), PostgresShardGeneratedIDField), cls._meta.fields))
        if postgres_shard_id_fields:
            database_dicts = [settings.DATABASES[database]] if database else [db_settings for db, db_settings in
                                                                              iteritems(settings.DATABASES) if
                                                                              db_settings["SHARD_GROUP"] == shard_group]
            if any([database_dict['ENGINE'] not in Backends.POSTGRES for database_dict in database_dicts]):
                raise ShardedModelInitializationException(
                    'You cannot use a PostgresShardGeneratedIDField on a non-Postgres database.')

            register_migration_signal_for_model_receiver(apps.get_app_config(cls._meta.app_label),
                                                         PostgresShardGeneratedIDField.migration_receiver,
                                                         dispatch_uid=PRE_MIGRATION_DISPATCH_UID % cls._meta.app_label)

        if shard_group:
            sharded_fields = list(filter(lambda field: issubclass(type(field), ShardedIDFieldMixin), cls._meta.fields))
            if not sharded_fields and not postgres_shard_id_fields:
                raise ShardedModelInitializationException('All sharded models require a ShardedIDFieldMixin or a '
                                                          'PostgresShardGeneratedIDField.')

            if not list(filter(lambda field: field == cls._meta.pk, sharded_fields)) and not postgres_shard_id_fields:
                raise ShardedModelInitializationException('All sharded models require the ShardedAutoIDField or '
                                                          'PostgresShardGeneratedIDFieldto be the primary key. Set '
                                                          'primary_key=True on the field.')

            if not callable(getattr(cls, 'get_shard', None)):
                raise ShardedModelInitializationException('You must define a get_shard method on the sharded model.')

            setattr(cls, 'django_sharding__shard_group', shard_group)
            setattr(cls, 'django_sharding__is_sharded', True)

            # If the sharded by field is set, we will make our custom manager the default manager.
            if sharded_by_field:
                try:
                    if not isinstance(cls.objects, ShardManager):
                        if type(cls.objects) == Manager:
                            cls.add_to_class('objects', ShardManager())
                            if django.VERSION < (1, 10):
                                cls._base_manager = cls.objects
                        else:
                            raise ShardedModelInitializationException('You must use the default Django model manager or'
                                                                      ' your custom manager must inherit from '
                                                                      '``ShardManager``')
                except AttributeError as e:
                    if cls._meta.abstract:
                        if django.VERSION < (1, 10):
                            managers = [x[2] for x in cls._meta.abstract_managers]
                        else:
                            managers = cls._meta.managers

                        if not len(managers) > 0:
                            cls.add_to_class('objects', ShardManager())
                        elif not any([isinstance(x, ShardManager) for x in managers]):
                            raise ShardedModelInitializationException(
                                'Please either do not specify a manager in your '
                                'abstract base class %s, or if you are using a '
                                'custom manager, your custom manager must '
                                'inherit from ``ShardManager``' % cls.__name__
                            )
                    else:
                        # If it gets to this point, the error is a Django error and not a library one. Pass it through.
                        raise e
                setattr(cls, 'django_sharding__sharded_by_field', sharded_by_field)
                if not callable(getattr(cls, 'get_shard_from_id', None)):
                    raise ShardedModelInitializationException('You must define a get_shard_from_id method on the '
                                                              'sharded model if you define a "sharded_by_field".')

        return cls
    return configure



from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from django_sharding_library.utils import create_postgres_global_sequence, create_postgres_shard_id_function, verify_postres_id_field_setup_correctly


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            '--database',
            action='store',
            dest='database',
            default=None,
            help='Nominates a database to create the sequence on. Defaults to "all" databases.',
            choices=['all'] + self.get_all_but_replica_dbs(),
        )
        parser.add_argument(
            '--sequence-name',
            action='store',
            dest='sequence_name',
            default='global_id_sequence',
            help='The name of the sequence to create and verify exists.',
        )
        parser.add_argument(
            '--reset-sequence', action='store_true', dest='reset_sequence', default=False,
            help="Reset the sequence if it exists.",
        )
        parser.add_argument(
            '--dry-run', action='store_true', dest='dry_run', default=False,
            help="Check they exist but make no changes to the database.",
        )

    def get_all_but_replica_dbs(self):
        return sorted(list(filter(
            lambda db: not settings.DATABASES[db].get('PRIMARY', None),
            settings.DATABASES.keys()
        )))

    def handle(self, *args, **options):
        if not options['database'] or options['database'] == 'all':
            databases = self.get_all_but_replica_dbs()
        elif options['database'] not in self.get_all_but_replica_dbs():
            raise CommandError('You must migrate an existing primary DB.')
        else:
            databases = [options['database']]

        has_errors = False

        for database in databases:
            sequence_name = options["sequence_name"]
            try:
                shard_id = settings.DATABASES[database].get('SHARD_ID', None)
                if shard_id is None:
                    continue
                if not options["dry_run"]:
                    create_postgres_global_sequence(sequence_name=sequence_name, db_alias=database, reset_sequence=options["reset_sequence"])
                    create_postgres_shard_id_function(sequence_name=sequence_name, db_alias=database, shard_id=shard_id)
                if not verify_postres_id_field_setup_correctly(sequence_name=sequence_name, db_alias=database, function_name="next_sharded_id"):
                    raise Exception("The sequence could not be found.")
                self.stdout.write(getattr(self.style, "SUCCESS", lambda a: a)("\nDatabase {} with shard id {}:sequence {} is present").format(database, shard_id, sequence_name))
            except Exception as e:
                self.stdout.write(getattr(self.style, "ERROR", lambda a: a)("\nDatabase {}: Error occured: {}").format(database, str(e)))
                has_errors = True

        if has_errors:
            raise CommandError("Some databases do not have the sequence on them.")



from mock import patch
import unittest

from django.conf import settings
from django.db import connections
from django.db.utils import DataError, IntegrityError
from django.test import TestCase
from django.utils.six.moves import xrange

import time
from datetime import datetime

from django_sharding_library.constants import Backends
from django_sharding_library.fields import (
    ShardedIDFieldMixin,
    ShardLocalStorageFieldMixin,
    ShardStorageFieldMixin,
    ShardForeignKeyStorageFieldMixin,
    ShardForeignKeyStorageField,
)
from django_sharding_library.id_generation_strategies import BaseIDGenerationStrategy
from tests.models import (
    ShardedModelIDs,
    ShardedTestModelIDs,
    TestModel,
    ShardStorageTable,
    PostgresCustomIDModel,
    PostgresShardUser
)


class BigAutoFieldTestCase(TestCase):
    def test_largest_id(self):
        if settings.DATABASES['default']['ENGINE'] in Backends.POSTGRES + Backends.SQLITE:
            max_id = 9223372036854775807
        else:
            max_id = 18446744073709551615

        item = ShardedModelIDs.objects.create(id=max_id, stub=None)
        self.assertEqual(item.pk, max_id)
        try:
            ShardedModelIDs.objects.create(id=max_id + 1, stub=None)
        except (DataError, OverflowError, IntegrityError):
            pass
        else:
            raise Exception('Field was able to hold an integer as large as {}.'.format(max_id + 1))


class FakeIDGenerationStrategy(BaseIDGenerationStrategy):
    def __init__(self):
        self.call_count = 0
        self.id = 0

    def get_next_id(self):
        self.call_count += 1
        self.id += 1
        return self.id


class TableShardedIDFieldTestCase(TestCase):
    def test_largest_id(self):
        if settings.DATABASES['app_shard_001']['ENGINE'] in Backends.POSTGRES + Backends.SQLITE:
            max_id = 9223372036854775807
        else:
            max_id = 18446744073709551615

        item = TestModel.objects.using('app_shard_001').create(id=max_id, user_pk=1)
        self.assertEqual(item.pk, max_id)
        try:
            TestModel.objects.using('app_shard_001').create(id=max_id + 1, user_pk=1)
        except (DataError, OverflowError, IntegrityError):
            pass
        else:
            raise Exception('Field was able to hold an integer as large as {}.'.format(max_id + 1))

    def test_get_pk_value_on_save_calls_strategy_get_next_id(self):
        for i in xrange(100):
            instance = TestModel.objects.using('app_shard_001').create(user_pk=1)
            self.assertEqual(ShardedTestModelIDs.objects.using('app_shard_001').latest('pk').pk, instance.pk)


class TableShardedIDMixinTestCase(TestCase):
    def test_initialization_makes_id_generation_available(self):
        field = ShardedIDFieldMixin(strategy=FakeIDGenerationStrategy())
        self.assertTrue(hasattr(field.strategy, 'get_next_id'))

    def test_uses_get_next_id_in_get_pk_value_on_save(self):
        class FakeModel(object):
            pk = None

        field = ShardedIDFieldMixin(strategy=FakeIDGenerationStrategy())
        for i in xrange(1, 100):
            self.assertEqual(field.get_pk_value_on_save(FakeModel()), i)
        self.assertEqual(field.strategy.id, 99)
        self.assertEqual(field.strategy.call_count, 99)


class ShardStorageFieldMixinTestCase(TestCase):
    def test_takes_shard_group_and_sets_attributes(self):
        sut = ShardStorageFieldMixin(shard_group='testing')
        self.assertEqual(sut.django_sharding__shard_group, 'testing')
        self.assertEqual(sut.django_sharding__stores_shard, True)
        self.assertFalse(hasattr(sut, 'django_sharding__use_signal'))


class ShardLocalStorageFieldMixinTestCase(TestCase):
    def test_takes_shard_group_and_sets_attributes(self):
        sut = ShardLocalStorageFieldMixin(shard_group='testing')
        self.assertEqual(sut.django_sharding__shard_group, 'testing')
        self.assertEqual(sut.django_sharding__stores_shard, True)
        self.assertEqual(sut.django_sharding__use_signal, True)


class ShardForeignKeyStorageFieldMixinTestCase(TestCase):
    def test_save_shard_works(self):
        initial_count = ShardStorageTable.objects.count()

        class FakeField(ShardForeignKeyStorageFieldMixin):
            def __init__(self, *args, **kwargs):
                self.name = 'shard_field'
                self.django_sharding__shard_storage_table = ShardStorageTable
                self.django_sharding__shard_group = 'default'

        class FakeModel(object):
            def __init__(self, shard_key):
                self.shard_key = shard_key
                self.shard_field = None

            def get_shard_key(self):
                return self.shard_key

        sut = FakeField(to=ShardStorageTable, shard_group='default')
        model = FakeModel('testing_save_shard')
        sut.save_shard(model)

        latest_entry_in_storage_table = ShardStorageTable.objects.latest('pk')
        self.assertEqual(model.shard_field, latest_entry_in_storage_table)
        self.assertEqual(ShardStorageTable.objects.count(), initial_count + 1)
        self.assertTrue(ShardStorageTable.objects.filter(shard_key='testing_save_shard'))
        self.assertTrue(latest_entry_in_storage_table.shard in settings.DATABASES)

    def test_save_shard_works_when_model_specified_by_string(self):
        initial_count = ShardStorageTable.objects.count()

        class FakeField(ShardForeignKeyStorageFieldMixin):
            def __init__(self, *args, **kwargs):
                self.name = 'shard_field'
                self.django_sharding__shard_storage_table = ShardStorageTable
                self.django_sharding__shard_group = 'default'

        class FakeModel(object):
            def __init__(self, shard_key):
                self.shard_key = shard_key
                self.shard_field = None

            def get_shard_key(self):
                return self.shard_key

        sut = FakeField(to='tests.ShardStorageTable', shard_group='default')
        model = FakeModel('testing_save_shard')
        sut.save_shard(model)

        latest_entry_in_storage_table = ShardStorageTable.objects.latest('pk')
        self.assertEqual(model.shard_field, latest_entry_in_storage_table)
        self.assertEqual(ShardStorageTable.objects.count(), initial_count + 1)
        self.assertTrue(ShardStorageTable.objects.filter(shard_key='testing_save_shard'))
        self.assertTrue(latest_entry_in_storage_table.shard in settings.DATABASES)


class ShardForeignKeyStorageFieldTestCase(TestCase):
    def test_pre_save_calls_save_shard(self):
        sut = ShardForeignKeyStorageField(ShardStorageTable, shard_group='default')
        model_instance = object()
        with patch.object(sut, 'save_shard') as mock_save_shard:
            with self.assertRaises(Exception):
                sut.pre_save(model_instance, False)

        mock_save_shard.assert_called_once_with(model_instance)


class PostgresShardIdFieldTestCase(TestCase):

    @unittest.skipIf(settings.DATABASES['default']['ENGINE'] not in Backends.POSTGRES, "Not a postgres backend")
    def test_check_shard_id_function(self):
        cursor = connections['default'].cursor()
        cursor.execute("SELECT next_sharded_id();")
        generated_id = cursor.fetchone()
        cursor.close()

        # Lets produce an ID that would have been generated 10 seconds ago on shard id 0 and assuming the counter was
        # at 0 (basically, the ID should be higher than this at all times, since time is the main factor)
        lowest_id = int(time.mktime(datetime.now().timetuple()) * 1000) - settings.SHARD_EPOCH - 10000 << 23
        lowest_id |= 0 << 10
        lowest_id |= 1

        self.assertGreater(generated_id[0], lowest_id)

    @unittest.skipIf(settings.DATABASES['default']['ENGINE'] not in Backends.POSTGRES, "Not a postgres backend")
    def test_check_shard_id_returns_with_model_save(self):
        user = PostgresShardUser.objects.create_user(username='username', password='pwassword', email='test@example.com')
        created_model = PostgresCustomIDModel.objects.create(random_string='Test String', user_pk=user.id)
        self.assertTrue(getattr(created_model, 'id'))

        # Same as above, lets create an id that would have been made 10 seconds ago and make sure the one that was
        # created and returned is larger
        lowest_id = int(time.mktime(datetime.now().timetuple()) * 1000) - settings.SHARD_EPOCH - 10000 << 23
        lowest_id |= 0 << 10
        lowest_id |= 1
        self.assertGreater(created_model.id, lowest_id)



postgres_shard_id_function_sql = """CREATE OR REPLACE FUNCTION next_sharded_id(OUT result bigint) AS $$
DECLARE
    start_epoch bigint := %(shard_epoch)d;
    seq_id bigint;
    now_millis bigint;
    shard_id int := %(shard_id)d;
BEGIN
    -- there is a typo here in the online example, which is corrected here
    SELECT nextval('%(sequence_name)s') %% 1024 INTO seq_id;

    SELECT FLOOR(EXTRACT(EPOCH FROM clock_timestamp()) * 1000) INTO now_millis;
    result := (now_millis - start_epoch) << 23;
    result := result | (shard_id << 10);
    result := result | (seq_id);
END;
$$ LANGUAGE PLPGSQL;"""



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


def forwards_func(apps, schema_editor):
    pass


def reverse_func(apps, schema_editor):
    pass


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0003_unshardedtestmodel_more_random_string'),
    ]

    operations = [
        migrations.RunPython(forwards_func, reverse_func, hints={"model_name": "tests.TestModel"}),
    ]



from django.conf import settings
from django.core.management.commands.migrate import Command as MigrationCommand

from django_sharding_library.exceptions import InvalidMigrationException


class Command(MigrationCommand):
    def handle(self, *args, **options):
        if not options['database'] or options['database'] == 'all':
            databases = self.get_all_but_replica_dbs()
        elif options['database'] not in self.get_all_but_replica_dbs():
            raise InvalidMigrationException('You must migrate an existing non-primary DB.')
        else:
            databases = [options['database']]

        for database in databases:
            options['database'] = database
            # Writen in green text to stand out from the surrouding headings
            if options['verbosity'] >= 1:
                self.stdout.write(getattr(self.style, "MIGRATE_SUCCESS", getattr(self.style, "SUCCESS", lambda a: a))("\nDatabase: {}\n").format(database))
            super(Command, self).handle(*args, **options)

    def get_all_but_replica_dbs(self):
        return list(filter(
            lambda db: not settings.DATABASES[db].get('PRIMARY', None),
            settings.DATABASES.keys()
        ))

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser._option_string_actions['--database'].default = None
        parser._option_string_actions['--database'].help = u'Nominates a database to synchronize. Defaults to all databases.'
        parser._option_string_actions['--database'].choices = ['all'] + self.get_all_but_replica_dbs()



from django.apps import AppConfig, apps
from django.conf import settings
from django.db import models
from django.dispatch import receiver

from django_sharding_library.routing_read_strategies import PrimaryOnlyRoutingStrategy
from django_sharding_library.sharding_functions import RoundRobinBucketingStrategy
from django_sharding_library.signals import save_shard_handler


class ShardingConfig(AppConfig):
    name = 'django_sharding'

    def ready(self):
        shard_settings = getattr(settings, 'DJANGO_SHARDING_SETTINGS', {})
        shard_groups = [settings.DATABASES[db_settings]['SHARD_GROUP'] for db_settings in settings.DATABASES]
        shard_groups = set(filter(lambda group: group is not None, shard_groups))
        self.bucketers = {}
        self.routing_strategies = {}
        for shard_group in shard_groups:
            group_settings = shard_settings.get(shard_group, {})
            self.bucketers[shard_group] = group_settings.get(
                'BUCKETING_STRATEGY',
                RoundRobinBucketingStrategy(shard_group=shard_group, databases=settings.DATABASES)
            )
            self.routing_strategies[shard_group] = group_settings.get(
                'ROUTING_STRATEGY',
                PrimaryOnlyRoutingStrategy(databases=settings.DATABASES)
            )

        # Unless otherwise instructed, add the signal to save the shard to the model if it has a shard field.
        for model in apps.get_models():
            if getattr(model, 'django_sharding__stores_shard', False) and getattr(model, 'django_sharding__shard_group', None):
                shard_group = getattr(model, 'django_sharding__shard_group', None)
                shard_field = getattr(model, 'django_sharding__shard_field', None)
                if not shard_field:
                    raise Exception('The model {} must have a `shard_field` attribute'.format(model))
            else:
                shard_fields = list(filter(lambda field: getattr(field, 'django_sharding__stores_shard', False), model._meta.fields))
                if not any(shard_fields):
                    continue

                if len(shard_fields) > 1:
                    raise Exception('The model {} has multuple fields for shard storage: {}'.format(model, shard_fields))
                shard_field = shard_fields[0]
                shard_group = getattr(shard_field, 'django_sharding__shard_group', None)

                if not shard_group:
                    raise Exception('The model {} with the shard field must have a `shard_group` attribute'.format(model))

                if not getattr(shard_field, 'django_sharding__use_signal', False):
                    continue

            group_settings = shard_settings.get(shard_group, {})
            if group_settings.get('SKIP_ADD_SHARDED_SIGNAL', False):
                continue

            receiver(models.signals.pre_save, sender=model)(save_shard_handler)

    def get_routing_strategy(self, shard_group):
        return self.routing_strategies[shard_group]

    def get_bucketer(self, shard_group):
        return self.bucketers[shard_group]



from django.apps import apps
from django.conf import settings
from django.db.models import AutoField, CharField, ForeignKey, BigIntegerField, OneToOneField

from django_sharding_library.constants import Backends
from django_sharding_library.utils import create_postgres_global_sequence, create_postgres_shard_id_function

try:
    from django.db.backends.postgresql.base import DatabaseWrapper as PostgresDatabaseWrapper
except ImportError:
    from django.db.backends.postgresql_psycopg2.base import DatabaseWrapper as PostgresDatabaseWrapper

try:
    from django.db.models import BigAutoField
except ImportError:
    from django.utils.translation import ugettext_lazy as _

    class BigAutoField(AutoField):
        description = _("Big (8 byte) integer")

        def db_type(self, connection):
            if connection.settings_dict['ENGINE'] in Backends.MYSQL:
                return 'serial'
            if connection.settings_dict['ENGINE'] in Backends.POSTGRES:
                return 'bigserial'
            return super(BigAutoField, self).db_type(connection)

        def rel_db_type(self, connection):
            return BigIntegerField().db_type(connection=connection)


class ShardedIDFieldMixin(object):
    """
    An autoincrimenting field which takes an id generator class instance
    as an argument and uses the generator to assign each new object a unique
    id.
    Note: This currently must be the primary key of the model and although this
    may be updated in the future, it should not hinder the app to use other
    candidates as unique fields.
    """
    def __init__(self, *args, **kwargs):
        # Remove the strategy from the kwargs so that it doesn't get passed to Django.
        setattr(self, 'strategy', kwargs['strategy'])
        del kwargs['strategy']
        return super(ShardedIDFieldMixin, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(ShardedIDFieldMixin, self).deconstruct()

        # Add the strategy from the kwargs so that it does get passed to our model.
        kwargs['strategy'] = getattr(self, 'strategy')
        return name, path, args, kwargs

    def get_pk_value_on_save(self, instance):
        if not instance.pk:
            return self.strategy.get_next_id()
        return instance.pk


class TableShardedIDField(ShardedIDFieldMixin, BigAutoField):
    """
    An autoincrimenting field which takes a `source_table` as an argument in
    order to generate unqiue ids for the sharded model.
    """
    def __init__(self, *args, **kwargs):
        from django_sharding_library.id_generation_strategies import TableStrategy
        kwargs['strategy'] = TableStrategy(backing_model=kwargs['source_table'])
        setattr(self, 'source_table', kwargs['source_table'])
        del kwargs['source_table']
        return super(TableShardedIDField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(TableShardedIDField, self).deconstruct()
        kwargs['source_table'] = getattr(self, 'source_table')
        return name, path, args, kwargs


class ShardedUUID4Field(ShardedIDFieldMixin, CharField):
    def __init__(self, *args, **kwargs):
        from django_sharding_library.id_generation_strategies import UUIDStrategy
        kwargs['strategy'] = UUIDStrategy()
        return super(ShardedUUID4Field, self).__init__(*args, **kwargs)

    def get_pk_value_on_save(self, instance):
        return self.strategy.get_next_id(instance.get_shard())


class ShardStorageFieldMixin(object):
    """
    A mixin for a field used to store a shard for in an instance or parent of an instance.
    """
    def __init__(self, *args, **kwargs):
        setattr(self, 'django_sharding__stores_shard', True)
        setattr(self, 'django_sharding__shard_group', kwargs['shard_group'])
        del kwargs['shard_group']
        return super(ShardStorageFieldMixin, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(ShardStorageFieldMixin, self).deconstruct()
        kwargs['shard_group'] = getattr(self, 'django_sharding__shard_group')
        return name, path, args, kwargs


class ShardLocalStorageFieldMixin(ShardStorageFieldMixin):
    """
    The ShardLocalStorageFieldMixin is used for when the shard is stored on the model that
    is being sharded by. i.e. Storing the shard on the User model and sharding by the User.
    """
    def __init__(self, *args, **kwargs):
        setattr(self, 'django_sharding__use_signal', True)
        return super(ShardLocalStorageFieldMixin, self).__init__(*args, **kwargs)

    def deconstruct(self):
        return super(ShardLocalStorageFieldMixin, self).deconstruct()


class ShardStorageCharField(ShardLocalStorageFieldMixin, CharField):
    """
    A simple char field that stores a shard and uses a signal to generate
    the shard using a pre_save signal.
    """
    pass


class ShardForeignKeyStorageFieldMixin(ShardStorageFieldMixin):
    """
    A mixin for a field used to store a foreign key to another table which
    stores the shard, usually a table which inherits from the ShardStorageModel.
    """
    def __init__(self, *args, **kwargs):
        setattr(self, 'django_sharding__stores_shard', True)
        model_class = kwargs.get('to', args and args[0])
        if type(model_class) == str:
            app_label = model_class.split('.')[0]
            app = apps.get_app_config(app_label)
            model_class = app.get_model(model_class[len(app_label) + 1:])
        setattr(self, 'django_sharding__shard_storage_table', model_class)
        return super(ShardForeignKeyStorageFieldMixin, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(ShardForeignKeyStorageFieldMixin, self).deconstruct()
        return name, path, args, kwargs

    def pre_save(self, model_instance, add):
        self.save_shard(model_instance)
        return super(ShardForeignKeyStorageFieldMixin, self).pre_save(model_instance, add)

    def save_shard(self, model_instance):
        shard_key = model_instance.get_shard_key()
        if not getattr(model_instance, self.name):
            shard_storage_table = getattr(self, 'django_sharding__shard_storage_table')
            shard_group = getattr(self, 'django_sharding__shard_group')

            app_config_app_label = getattr(settings, 'DJANGO_SHARDING_SETTINGS', {}).get('APP_CONFIG_APP', 'django_sharding')
            bucketer = apps.get_app_config(app_config_app_label).get_bucketer(shard_group)
            shard = bucketer.pick_shard(model_instance)
            shard_object, _ = shard_storage_table.objects.get_or_create(shard_key=shard_key)
            if not shard_object.shard:
                shard_object.shard = shard
                shard_object.save()
            setattr(model_instance, self.name, shard_object)


class ShardForeignKeyStorageField(ShardForeignKeyStorageFieldMixin, ForeignKey):
    """
    A simple char field that stores a shard and uses a signal to generate
    the shard using a pre_save signal.
    """
    pass


class PostgresShardGeneratedIDField(AutoField):
    """
    A field that uses a Postgres stored procedure to return an ID generated on the database.
    """
    def db_type(self, connection, *args, **kwargs):

        if not hasattr(settings, 'SHARD_EPOCH'):
            raise ValueError("PostgresShardGeneratedIDField requires a SHARD_EPOCH to be defined in your settings file.")

        if connection.vendor == PostgresDatabaseWrapper.vendor:
            return "bigint DEFAULT next_sharded_id()"
        else:
            return super(PostgresShardGeneratedIDField, self).db_type(connection)

    def get_internal_type(self):
        return 'BigIntegerField'

    def rel_db_type(self, connection):
        return BigIntegerField().db_type(connection=connection)

    @staticmethod
    def migration_receiver(*args, **kwargs):
        sequence_name = "global_id_sequence"
        db_alias = kwargs.get('using')
        if not db_alias:
            raise EnvironmentError("A pre-migration receiver did not receive a database alias. "
                                   "Perhaps your app is not registered correctly?")
        if settings.DATABASES[db_alias]['ENGINE'] in Backends.POSTGRES:
            shard_id = settings.DATABASES[db_alias].get('SHARD_ID', 0)
            create_postgres_global_sequence(sequence_name, db_alias, True)
            create_postgres_shard_id_function(sequence_name, db_alias, shard_id)


class PostgresShardForeignKey(ForeignKey):
    def db_type(self, connection):
        # The database column type of a ForeignKey is the column type
        # of the field to which it points. An exception is if the ForeignKey
        # points to an AutoField/PositiveIntegerField/PositiveSmallIntegerField,
        # in which case the column type is simply that of an IntegerField.
        # If the database needs similar types for key fields however, the only
        # thing we can do is making AutoField an IntegerField.
        rel_field = self.target_field
        if rel_field.get_internal_type() == "BigIntegerField":
            return BigIntegerField().db_type(connection=connection)
        return super(PostgresShardForeignKey, self).db_type(connection)


class PostgresShardOneToOne(OneToOneField):
    def db_type(self, connection):
        # The database column type of a ForeignKey is the column type
        # of the field to which it points. An exception is if the ForeignKey
        # points to an AutoField/PositiveIntegerField/PositiveSmallIntegerField,
        # in which case the column type is simply that of an IntegerField.
        # If the database needs similar types for key fields however, the only
        # thing we can do is making AutoField an IntegerField.
        rel_field = self.target_field
        if rel_field.get_internal_type() == "BigIntegerField":
            return BigIntegerField().db_type(connection=connection)
        return super(PostgresShardOneToOne, self).db_type(connection)



import uuid

from django.db import connections, transaction
from django.utils.deconstruct import deconstructible

from django_sharding_library.constants import Backends
from django_sharding_library.models import TableStrategyModel


class BaseIDGenerationStrategy(object):
    """
    A strategy for Generating unique identifiers for the sharded models.
    """
    def get_next_id(self, database=None):
        """
        A function which takes no arguments and returns a new unique identifier.
        """
        raise NotImplemented


@deconstructible
class TableStrategy(BaseIDGenerationStrategy):
    """
    Uses an autoincrement field, on a TableStrategyModel model `backing_model`
    to generate unique IDs.
    """
    def __init__(self, backing_model):
        if not issubclass(backing_model, TableStrategyModel):
            raise ValueError("Unsupported model used for generating IDs")
        self.backing_model = backing_model

    def get_next_id(self, database=None):
        """
        Returns a new unique integer identifier for an object using an
        auto-incrimenting field in the database.
        """
        from django.conf import settings
        backing_table_db = getattr(self.backing_model, 'database', 'default')
        if settings.DATABASES[backing_table_db]['ENGINE'] in Backends.MYSQL:
            with transaction.atomic(backing_table_db):
                cursor = connections[backing_table_db].cursor()
                sql = "REPLACE INTO `{0}` (`stub`) VALUES ({1})".format(
                    self.backing_model._meta.db_table, True
                )
                cursor.execute(sql)

            if getattr(cursor.cursor.cursor, 'lastrowid', None):
                id = cursor.cursor.cursor.lastrowid
            else:
                id = self.backing_model.objects.get(stub=True).id
        else:
            with transaction.atomic(backing_table_db):
                id = self.backing_model.objects.create(stub=None).id
        return id


@deconstructible
class UUIDStrategy(BaseIDGenerationStrategy):
    """
    Uses a uuid and the shard name to generate unique IDs.
    """
    def get_next_id(self, database):
        """
        Generate a cross-shard UUID using a python UUID and the name of the database.
        """
        from django.conf import settings
        assert database in settings.DATABASES
        return '{}-{}'.format(database, uuid.uuid4())



from unittest import skip

from django.core.management import call_command
from django.test import TestCase
from mock import patch

from django_sharding_library.exceptions import InvalidMigrationException


@patch('django.core.management.commands.migrate.Command.handle')
class MigrationCommandTestCase(TestCase):

    def test_defauls_migrates_all_primary_dbs(self, mock_migrate_command):
        call_command('migrate', verbosity=0)
        databases_migrated = [call[1].get('database') for call in mock_migrate_command.call_args_list]
        expected_migrated_databases = ['app_shard_001', 'app_shard_002', 'app_shard_003', 'app_shard_004', 'default']

        self.assertEqual(sorted(databases_migrated), expected_migrated_databases)

    def test_all_option_added_to_databases(self, mock_migrate_command):
        call_command('migrate', database='all', verbosity=0)
        databases_migrated = [call[1].get('database') for call in mock_migrate_command.call_args_list]
        expected_migrated_databases = ['app_shard_001', 'app_shard_002', 'app_shard_003', 'app_shard_004', 'default']

        self.assertEqual(sorted(databases_migrated), expected_migrated_databases)

    def test_migrate_single_db(self, mock_migrate_command):
        call_command('migrate', database='default', verbosity=0)
        databases_migrated = [call[1].get('database') for call in mock_migrate_command.call_args_list]
        expected_migrated_databases = ['default']

        self.assertEqual(sorted(databases_migrated), expected_migrated_databases)

    def test_migrate_replica_raises_exception(self, mock_migrate_command):
        with self.assertRaises(InvalidMigrationException):
            call_command('migrate', database='app_shard_001_replica_001', verbosity=0)

        databases_migrated = []
        expected_migrated_databases = []

        self.assertEqual(databases_migrated, expected_migrated_databases)

    @skip("skipped until fix https://code.djangoproject.com/ticket/26597")
    def test_passes_other_args(self, mock_migrate_command):
        call_command('migrate', database='app_shard_001', fake=True, verbosity=0, app_label='tests', migration_name='0001', noinput=True, fake_initial=True, interactive=False, list=True)

        self.assertTrue(mock_migrate_command.call_args[1].get('fake'))
        self.assertTrue(mock_migrate_command.call_args[1].get('fake_initial'))
        self.assertFalse(mock_migrate_command.call_args[1].get('interactive'))
        self.assertTrue(mock_migrate_command.call_args[1].get('list'))
        self.assertEqual(mock_migrate_command.call_args[1].get('verbosity'), 0)
        self.assertEqual(mock_migrate_command.call_args[1].get('database'), 'app_shard_001')
        self.assertEqual(mock_migrate_command.call_args[1].get('app_label'), 'tests')
        self.assertEqual(mock_migrate_command.call_args[1].get('migration_name'), '0001')



from dj_database_url import config


def database_config(environment_variable, default_database_url, shard_group=None, is_replica_of=None):
    """
    Wraps dj_database_url to provide additional arguments to specify whether a database is a shard
    and if it a replica of another database.
    """
    db_config = config(env=environment_variable, default=default_database_url)
    if not db_config:
        return db_config

    db_config['TEST'] = db_config.get('TEST', {})
    db_config['SHARD_GROUP'] = shard_group

    if is_replica_of:
        db_config['PRIMARY'] = is_replica_of
        db_config['TEST']['MIRROR'] = is_replica_of

    return db_config


def database_configs(databases_dict):
    """
    Takes databases of the form:
    {
        'unsharded_databases': [
            {
                'name': 'DB01',
                'environment_variable': 'ENV',
                'default_database_url': 'postgres:://...'
            }, {
                'name': 'DB02',
               'environment_variable': 'ENV2',
                'default_database_url': 'postgres:://...',
                'replicas': [{
                    'name': 'DB02_S1',
                    'environment_variable': 'ENVS1',
                    'default_database_url': 'postgres:://...'
                }, {
                    'name': 'DB02_S2',
                    'environment_variable': 'ENVS2',
                    'default_database_url': 'postgres:://...'
                }]
            },],
        'sharded_databases': [
            {
                'name': 'SHARD_01',
                'environment_variable': 'ENV3',
                'default_database_url': 'postgres:://...',
                'shard_group': 'default',
                'replicas': [{
                    'name': 'REPLICA_SHARD_01_01',
                    'environment_variable': 'ENVS3',
                    'default_database_url': 'postgres:://...'
                    'shard_group': 'default',
                }, {
                    'name': 'REPLICA_SHARD_01_02',
                    'environment_variable': 'ENVS4',
                    'default_database_url': 'postgres:://...'
                    'shard_group': 'default',
                }]
            }, {
                'name': 'SHARD_02',
                'environment_variable': 'ENV4',
                'default_database_url': 'postgres:://...'
                'shard_group': 'default',
            },
        ]
    }
    """
    configuration = {}
    shard_id_hash = {}  # Keep track of the IDs of the shards currently. Used to help with migrations.
    for (databases, is_sharded) in [(databases_dict.get('unsharded_databases', []), False), (databases_dict.get('sharded_databases', []), True)]:
        for idx, database in enumerate(databases):
            db_config = database_config(
                database['environment_variable'],
                database['default_database_url'],
                shard_group=(is_sharded and database.get('shard_group', 'default')) or None,
                is_replica_of=None
            )
            if db_config:
                configuration[database['name']] = db_config
            for replica in database.get('replicas', []):
                db_config = database_config(
                    replica['environment_variable'],
                    replica['default_database_url'],
                    shard_group=(is_sharded and database.get('shard_group', 'default') or None),
                    is_replica_of=database['name']
                )
                if db_config:
                    configuration[replica['name']] = db_config

            # We assume the numeric shard ID is constant based on the entries in the configuration helper (we assume
            # they wont change order, and that new shards will be appended and not inserted randomly)
            # This is noted in the docs, leaving this comment for whomever may work on this in the future.
            if is_sharded:
                shard_id = shard_id_hash.get(configuration[database['name']]['SHARD_GROUP'], 0)
                configuration[database['name']]['SHARD_ID'] = shard_id
                shard_id_hash[configuration[database['name']]['SHARD_GROUP']] = shard_id + 1

    return configuration



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-08 18:38
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0004_remove_unshardedtestmodel_more_random_string'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='unshardedtestmodel',
            name='more_random_string',
        ),
    ]



from django.conf import settings
from django.db import models

from django_sharding_library.fields import BigAutoField


def _get_primary_shards():
    """
    Returns the names of databases which make up the shards and have no primary.
    """
    return sorted(
        filter(
            lambda db: not settings.DATABASES[db].get('PRIMARY', None) and settings.DATABASES[db].get('SHARD_GROUP', None),
            settings.DATABASES.keys()
        )
    )


class ShardedByMixin(models.Model):
    """
    This mixin is intended to be used with the included `save_shard_handler` in order
    to save the shard to the object. This is done by registering it on the User's pre-save signal.
    Look at the signal's docstring for additional information.

    Note: requires the use of `cls.shard_group`
    """
    django_sharding__shard_field = 'shard'
    django_sharding__stores_shard = True

    SHARD_CHOICES = ((i, i) for i in _get_primary_shards())

    shard = models.CharField(max_length=120, blank=True, null=True, choices=SHARD_CHOICES)

    class Meta:
        abstract = True


class TableStrategyModel(models.Model):
    """
    A model for use with the TableStrategy in order to generate ids using an
    autoincrimenting field. To use this, simply inherit from this class and give
    the model a more meaningful name e.g. InvoiceIDs.
    """
    id = BigAutoField(primary_key=True)
    stub = models.NullBooleanField(null=True, default=True, unique=True)

    class Meta:
        abstract = True


class ShardStorageModel(models.Model):
    """
    A model for storing shards.
    """
    SHARD_CHOICES = ((i, i) for i in _get_primary_shards())

    shard = models.CharField(max_length=120, choices=SHARD_CHOICES)
    shard_key = models.CharField(primary_key=True, max_length=120)

    class Meta:
        abstract = True



from django.apps import apps
from django.conf import settings


def save_shard_handler(sender, instance, **kwargs):
    """
    Saves the shard to the model in the field `shard`.
    e.g. usage:
    @receiver(models.signals.pre_save, sender=User)
    def shard_handler(sender, instance, **kwargs):
        save_shard_handler(sender, instance, **kwargs)
    """
    app_config_app_label = getattr(settings, 'DJANGO_SHARDING_SETTINGS', {}).get('APP_CONFIG_APP', 'django_sharding')
    bucketer = apps.get_app_config(app_config_app_label).get_bucketer(sender.django_sharding__shard_group)
    shard_fields = list(filter(lambda field: getattr(field, 'django_sharding__stores_shard', False), sender._meta.fields))

    if not any(shard_fields):
        shard_field_name = getattr(sender, 'django_sharding__shard_field', None)
        shard_fields = list(filter(lambda field: field.name == shard_field_name, sender._meta.fields))

    if not any(shard_fields):
        return

    if len(shard_fields) > 1:
        raise Exception('The model {} has multuple fields for shard storage: {}'.format(sender, shard_fields))
    shard_field = shard_fields[0]

    if not getattr(instance, shard_field.name, None):
        setattr(instance, shard_field.name, bucketer.pick_shard(instance))



from django.apps import apps
from django.conf import settings

from django_sharding_library.fields import PostgresShardGeneratedIDField
from django_sharding_library.exceptions import DjangoShardingException, InvalidMigrationException
from django_sharding_library.utils import (
    is_model_class_on_database,
    get_database_for_model_instance,
    get_possible_databases_for_model,
)


class ShardedRouter(object):
    """
    A router that is shard-aware and will prevent running migrations on
    the wrong database as well as infer, when possible, which database to
    read or write from.
    """

    def get_shard_for_instance(self, instance):
        return instance._state.db or instance.get_shard()

    def get_shard_for_id_field(self, model, sharded_by_field_id):
        try:
            return model.get_shard_from_id(sharded_by_field_id)
        except Exception:
            # This is overly broad on purpose: we dont know what a user might try to do to get a shard from an ID (our
            # example is a .get(), but someone could do anything) and if it excepts, we want to fall back to the next
            # attempt at grabbing a shard based on other potential filters
            return None

    def get_shard_for_postgres_pk_field(self, model, pk_value):
        group = getattr(model, 'django_sharding__shard_group', None)
        shard_id_to_find = int(bin(pk_value)[-23:-10], 2)  # We know where the shard id is stored in the PK's bits.

        # We can check the shard id from the PK against the shard ID in the databases config
        for alias, db_settings in settings.DATABASES.items():
            if db_settings["SHARD_GROUP"] == group and db_settings["SHARD_ID"] == shard_id_to_find:
                return alias

        return None  # Return None if we could not determine the shard so we can fall through to the next shard grab attempt

    def get_read_db_routing_strategy(self, shard_group):
        app_config_app_label = getattr(settings, 'DJANGO_SHARDING_SETTINGS', {}).get('APP_CONFIG_APP', 'django_sharding')
        return apps.get_app_config(app_config_app_label).get_routing_strategy(shard_group)

    def _get_shard(self, model, **hints):
        shard = None

        model_has_sharded_id_field = getattr(model, 'django_sharding__sharded_by_field', None) is not None

        #####
        #
        # This is setup as multiple IF statements on purpose. If any attempt to get a shard fails, the function that
        # tried to get the shard should return None that way the NEXT possible attempt to grab a shard can be run.
        # In this way, future ways to automatically get the correct shard can be added, and at any point in the logic,
        # and the first one to pick a valid shard will return the valid shard.
        #
        # Always return None if no valid shard was found so the default Django router will use the using() database if
        # these all fail.
        #
        #####
        if hints.get("instance", None):
            shard = get_database_for_model_instance(instance=hints["instance"])
        if shard is None and model_has_sharded_id_field:
            sharded_by_field_id = hints.get('exact_lookups', {}).get(
                getattr(model, 'django_sharding__sharded_by_field'), None
            )
            if sharded_by_field_id:
                shard = self.get_shard_for_id_field(model, sharded_by_field_id)

        is_pk_postgres_generated_id_field = isinstance(getattr(model._meta, 'pk'), PostgresShardGeneratedIDField)
        lookup_pk = hints.get('exact_lookups', {}).get('pk') or hints.get('exact_lookups', {}).get('id')

        if shard is None and is_pk_postgres_generated_id_field and lookup_pk is not None:
            return self.get_shard_for_postgres_pk_field(model, lookup_pk)

        return shard

    def db_for_read(self, model, **hints):
        possible_databases = get_possible_databases_for_model(model=model)
        if len(possible_databases) == 1:
            return possible_databases[0]

        shard = self._get_shard(model, **hints)
        if shard:
            shard_group = getattr(model, 'django_sharding__shard_group', None)
            if not shard_group:
                raise DjangoShardingException('Unable to identify the shard_group for the {} model'.format(model))
            routing_strategy = self.get_read_db_routing_strategy(shard_group)
            return routing_strategy.pick_read_db(shard)
        return None

    def db_for_write(self, model, **hints):
        possible_databases = get_possible_databases_for_model(model=model)
        if len(possible_databases) == 1:
            return possible_databases[0]

        shard = self._get_shard(model, **hints)

        if shard:
            db_config = settings.DATABASES[shard]
            return db_config.get('PRIMARY', shard)
        return None

    def allow_relation(self, obj1, obj2, **hints):
        """
        Only allow relationships between two items which are both on only one database or
        between sharded items on the same shard.
        """

        object1_databases = get_possible_databases_for_model(model=obj1._meta.model)
        object2_databases = get_possible_databases_for_model(model=obj2._meta.model)

        if (len(object1_databases) == len(object2_databases) == 1) and (object1_databases == object2_databases):
            return True
        return self.get_shard_for_instance(obj1) == self.get_shard_for_instance(obj2)

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if settings.DATABASES[db].get('PRIMARY', None):
            return False

        # Since the API for this function is limiting in a sharded environemnt,
        # we provide an override to specify which databases to run the migrations
        # on.
        if hints.get("force_migrate_on_databases", None):
            return db in hints["force_migrate_on_databases"]

        model_name = model_name or hints.get('model_name')
        model = hints.get('model')
        if model:
            model_name = model.__name__

        # Return true if any model in the app is on this database.
        if not model_name:
            app = apps.get_app_config(app_label)
            for model in app.get_models():
                if is_model_class_on_database(model=model, database=db):
                    return True
            return False

        # Sometimes, when extending models from another app i.e. the User Model, the app label
        # is the app label of the app where the change is defined but to app with the model is
        # passed in with the model name.
        try:
            app = apps.get_app_config(app_label)
            model = app.get_model(model_name)
        except LookupError:
            app_label = model_name.split('.')[0]
            app = apps.get_app_config(app_label)
            model = app.get_model(model_name[len(app_label) + 1:])

        try:
            return is_model_class_on_database(model=model, database=db)
        except DjangoShardingException as e:
            raise InvalidMigrationException(
                e.args[0]
            )



class Backends(object):
    MYSQL = ('django.db.backends.mysql', 'django.contrib.gis.db.backends.mysql')
    POSTGRES = ('django.db.backends.postgresql_psycopg2', 'django.db.backends.postgresql', 'django.contrib.gis.db.backends.postgis')
    SQLITE = ('django.db.backends.sqlite3', 'django.contrib.gis.db.backends.spacialite')



from django.conf import settings
from django.test import TestCase
from django.utils.six.moves import xrange

from django_sharding_library.routing_read_strategies import RoundRobinRoutingStrategy, PrimaryOnlyRoutingStrategy


class RoundRobinBucketingStrategyTestCase(TestCase):

    def test_is_cyclic(self):
        sut = RoundRobinRoutingStrategy(settings.DATABASES)
        expected_cycled_shards = ['app_shard_001_replica_001', 'app_shard_001_replica_002', 'app_shard_001']
        expected_cycled_shards.sort()

        resulting_shards = [sut.pick_read_db('app_shard_001') for i in xrange(150)]

        self.assertEqual(len(set([resulting_shards[i] for i in xrange(0, 150, 3)])), 1)
        self.assertEqual(len(set([resulting_shards[i] for i in xrange(1, 150, 3)])), 1)
        self.assertEqual(len(set([resulting_shards[i] for i in xrange(2, 150, 3)])), 1)

        resulting_cycled_shard = resulting_shards[:3]
        resulting_cycled_shard.sort()

        self.assertEqual(expected_cycled_shards, resulting_cycled_shard)


class MasterOnlyRoutingStrategyTestCase(TestCase):

    def test_is_always_primary(self):
        sut = PrimaryOnlyRoutingStrategy(settings.DATABASES)
        expected_shards = ['app_shard_001'] * 150

        resulting_shards = [sut.pick_read_db('app_shard_001') for i in xrange(150)]
        self.assertEqual(expected_shards, resulting_shards)


class RandomRoutingStrategyTestCase(TestCase):

    def test_no_exception_raised(self):
        sut = RoundRobinRoutingStrategy(settings.DATABASES)
        [sut.pick_read_db('app_shard_001') for i in xrange(150)]



from django.core.management import call_command
from django.test import TestCase
from mock import call, patch


mock_create_sequence_path = 'django_sharding_library.management.commands.create_postgres_sequences.create_postgres_global_sequence'
mock_create_func_path = 'django_sharding_library.management.commands.create_postgres_sequences.create_postgres_shard_id_function'
mock_verify_path = 'django_sharding_library.management.commands.create_postgres_sequences.verify_postres_id_field_setup_correctly'


class MigrationCommandTestCase(TestCase):

    def test_defauls_migrates_all_primary_dbs_with_shard_id(self):
        with patch(mock_verify_path, return_value=True) as mock_verify:
            with patch(mock_create_func_path) as mock_create_func:
                with patch(mock_create_sequence_path) as mock_create_sequence:
                    call_command('create_postgres_sequences')

        mock_verify.assert_has_calls(
            [
                call(db_alias='app_shard_001', function_name='next_sharded_id', sequence_name='global_id_sequence'),
                call(db_alias='app_shard_002', function_name='next_sharded_id', sequence_name='global_id_sequence')
            ]
        )

        mock_create_func.assert_has_calls(
            [
                call(db_alias='app_shard_001', sequence_name='global_id_sequence', shard_id=0),
                call(db_alias='app_shard_002', sequence_name='global_id_sequence', shard_id=1)
            ]
        )

        mock_create_sequence.assert_has_calls(
            [
                call(db_alias='app_shard_001', reset_sequence=False, sequence_name='global_id_sequence'),
                call(db_alias='app_shard_002', reset_sequence=False, sequence_name='global_id_sequence')
            ]
        )

    def test_reset_sequences(self):
        with patch(mock_verify_path, return_value=True) as mock_verify:
            with patch(mock_create_func_path) as mock_create_func:
                with patch(mock_create_sequence_path) as mock_create_sequence:
                    call_command('create_postgres_sequences', '--reset-sequence')

        mock_verify.assert_has_calls(
            [
                call(db_alias='app_shard_001', function_name='next_sharded_id', sequence_name='global_id_sequence'),
                call(db_alias='app_shard_002', function_name='next_sharded_id', sequence_name='global_id_sequence')
            ]
        )

        mock_create_func.assert_has_calls(
            [
                call(db_alias='app_shard_001', sequence_name='global_id_sequence', shard_id=0),
                call(db_alias='app_shard_002', sequence_name='global_id_sequence', shard_id=1)
            ]
        )

        mock_create_sequence.assert_has_calls(
            [
                call(db_alias='app_shard_001', reset_sequence=True, sequence_name='global_id_sequence'),
                call(db_alias='app_shard_002', reset_sequence=True, sequence_name='global_id_sequence')
            ]
        )

    def test_specify_database(self):
        with patch(mock_verify_path, return_value=True) as mock_verify:
            with patch(mock_create_func_path) as mock_create_func:
                with patch(mock_create_sequence_path) as mock_create_sequence:
                    call_command('create_postgres_sequences', '--database=app_shard_001')

        mock_verify.assert_has_calls(
            [
                call(db_alias='app_shard_001', function_name='next_sharded_id', sequence_name='global_id_sequence'),
            ]
        )

        mock_create_func.assert_has_calls(
            [
                call(db_alias='app_shard_001', sequence_name='global_id_sequence', shard_id=0),
            ]
        )

        mock_create_sequence.assert_has_calls(
            [
                call(db_alias='app_shard_001', reset_sequence=False, sequence_name='global_id_sequence'),
            ]
        )

    def test_specify_sequence_name(self):
        with patch(mock_verify_path, return_value=True) as mock_verify:
            with patch(mock_create_func_path) as mock_create_func:
                with patch(mock_create_sequence_path) as mock_create_sequence:
                    call_command('create_postgres_sequences', '--sequence-name=test_sequence_name')

        mock_verify.assert_has_calls(
            [
                call(db_alias='app_shard_001', function_name='next_sharded_id', sequence_name='test_sequence_name'),
                call(db_alias='app_shard_002', function_name='next_sharded_id', sequence_name='test_sequence_name')
            ]
        )

        mock_create_func.assert_has_calls(
            [
                call(db_alias='app_shard_001', sequence_name='test_sequence_name', shard_id=0),
                call(db_alias='app_shard_002', sequence_name='test_sequence_name', shard_id=1)
            ]
        )

        mock_create_sequence.assert_has_calls(
            [
                call(db_alias='app_shard_001', reset_sequence=False, sequence_name='test_sequence_name'),
                call(db_alias='app_shard_002', reset_sequence=False, sequence_name='test_sequence_name')
            ]
        )

    def test_dry_run(self):
        with patch(mock_verify_path, return_value=True) as mock_verify:
            with patch(mock_create_func_path) as mock_create_func:
                with patch(mock_create_sequence_path) as mock_create_sequence:
                    call_command('create_postgres_sequences', '--dry-run')

        mock_verify.assert_has_calls(
            [
                call(db_alias='app_shard_001', function_name='next_sharded_id', sequence_name='global_id_sequence'),
                call(db_alias='app_shard_002', function_name='next_sharded_id', sequence_name='global_id_sequence')
            ]
        )

        mock_create_func.assert_has_calls(
            []
        )

        mock_create_sequence.assert_has_calls(
            []
        )



from django.db.models.manager import Manager
from django.db.models.query import QuerySet


class ShardQuerySet(QuerySet):
    """
    Stores the lookups to pass to the router as a hint
    """
    def __init__(self, model=None, *args, **kwargs):
        super(ShardQuerySet, self).__init__(model=model, *args, **kwargs)
        self._exact_lookups = {}

    def _clone(self, **kwargs):
        clone = super(ShardQuerySet, self)._clone(**kwargs)
        clone._exact_lookups = self._exact_lookups.copy()
        return clone

    def _filter_or_exclude(self, *args, **kwargs):
        """
        Update our lookups when we get a filter or an exclude (we only care about filter, but its a
        shared function in the ORM)
        :return:
        """
        clone = super(ShardQuerySet, self)._filter_or_exclude(*args, **kwargs)
        if getattr(clone, '_exact_lookups', None) is None:
            clone._exact_lookups = {}
        clone._exact_lookups.update(dict([(k, v) for k, v in kwargs.items() if '__' not in k]))
        return clone

    @property
    def db(self):
        if self._db:
            return self._db

        self._hints['exact_lookups'] = self._exact_lookups
        if not self._hints.get('instance') and getattr(self, '_instance', None):
            self._hints['instance'] = getattr(self, '_instance')

        return super(ShardQuerySet, self).db

    def create(self, **kwargs):
        """
        Grabs the instance before its too late to pass it to the router as a hint. Django (as of 1.9) does not keep
        instances around in all cases, such as a create() call.
        """
        self._instance = self.model(**kwargs.copy())
        return super(ShardQuerySet, self).create(**kwargs)

    def get_or_create(self, defaults=None, **kwargs):
        """
        Add the lookups to the _exact_lookups and call super.
        """
        defaults = defaults or {}
        lookup, params = self._extract_model_params(defaults, **kwargs)
        self._exact_lookups = lookup
        return super(ShardQuerySet, self).get_or_create(defaults=defaults, **kwargs)

    def update_or_create(self, defaults=None, **kwargs):
        """
        Add the lookups to the _exact_lookups and call super.
        """
        defaults = defaults or {}
        lookup, params = self._extract_model_params(defaults, **kwargs)
        self._exact_lookups = lookup
        return super(ShardQuerySet, self).update_or_create(defaults=defaults, **kwargs)


class ShardManager(Manager):

    def __init__(self, *args, **kwargs):
        return_value = super(ShardManager, self).__init__(*args, **kwargs)
        self.name = 'shard_manager'
        return return_value

    def get_queryset(self, key=None):
        # todo: (eventually, not necessary now) Should check to make sure the there is a good kwarg OR an instance ->
        # available here
        return ShardQuerySet(model=self.model)

    def _wrap(func_name):
        def wrapped(self, *args, **kwargs):
            return getattr(self.get_queryset(), func_name)(*args, **kwargs)

        wrapped.__name__ = func_name
        return wrapped

    filter = _wrap('filter')
    get = _wrap('get')
    create = _wrap('create')
    get_or_create = _wrap('get_or_create')
    update_or_create = _wrap('update_or_create')



from itertools import cycle
from random import choice, randint

from django.utils.six import next, viewitems


class BaseRoutingStrategy(object):
    """
    A base strategy for picking which database to read from when there are read
    replicas in the system. In order to extend this strategy, you must define a
    `pick_read_db` function which returns the name of the DB to read from,
    given a primary DB.
    If there are no read replicas defined, all strategies should always return the
    primary.
    """
    def __init__(self, databases):
        self.primary_replica_mapping = self.get_primary_replica_mapping(databases)

    def get_primary_replica_mapping(self, databases):
        """
        Creates a dictionary that maps a primary drive name to all the names of
        it's replication databases. This can be used in the strategies.
        """
        mapping = {}
        for name, config in databases.items():
            primary = config.get('PRIMARY', None)
            if not primary:
                continue
            if primary not in mapping:
                mapping[primary] = []
            if primary != name:
                mapping[primary].append(name)
        return mapping

    def pick_read_db(self, primary_db_name):
        """
        Given the name of a primary, pick the name of the database to read
        from which may be a replica or the primary itself.
        """
        raise NotImplemented


class PrimaryOnlyRoutingStrategy(BaseRoutingStrategy):
    """
    A strategy which will always read from the primary, unless overridden,
    regardless of replicas defined.
    """
    def pick_read_db(self, primary_db_name):
        return primary_db_name


class RoundRobinRoutingStrategy(BaseRoutingStrategy):
    """
    A strategy which will cycle through the read replicas and primary, in a
    round-robin fashion when choosing which database to read from.
    """
    def __init__(self, databases):
        super(RoundRobinRoutingStrategy, self).__init__(databases)
        self.read_cycles = {}

        for primary, replicas in viewitems(self.primary_replica_mapping):
            self.read_cycles[primary] = cycle(replicas + [primary])

    def pick_read_db(self, primary_db_name):
        return next(self.read_cycles[primary_db_name])


class RandomRoutingStrategy(BaseRoutingStrategy):
    """
    A strategy which will choose a random read replicas, or the primary,
    when choosing which database to read from.
    """
    def pick_read_db(self, primary_db_name):
        return choice(self.primary_replica_mapping[primary_db_name])


class RatioRoutingStrategy(BaseRoutingStrategy):
    def pick_read_db(self, primary_db_name):
        """
        Read from primary half the time and random replicas the other half.
        """
        if randint(0, 1):
            return primary_db_name
        return choice(self.primary_replica_mapping[primary_db_name])



from itertools import cycle
from random import choice, randint

from django.utils.six import next


class BaseBucketingStrategy(object):
    """
    A base strategy for bucketing Users into shards. In order to extend this
    strategy, you must define a `pick_shard` and a `get_shard` method that
    both take the model that is sharded by.
    """
    def __init__(self, shard_group='default'):
        self.shard_group = shard_group

    def get_shards(self, databases):
        shards = []
        for name, config in databases.items():
            if config.get('SHARD_GROUP') == self.shard_group and not config.get('PRIMARY'):
                shards.append(name)
        return shards

    def pick_shard(self, model_sharded_by):
        """
        Returns the shard for the model which has not previously been bucketed
        into a shard.
        """
        raise NotImplementedError

    def get_shard(self, model_sharded_by):
        """
        Returns the shard for a model which has already been assigned a shard.
        """
        raise NotImplementedError


class BaseShardedModelBucketingStrategy(BaseBucketingStrategy):
    """
    A base strategy for bucketing models into shards when saving the shard on
    the model.
    """
    def get_shard(self, model_sharded_by):
        sharded_field = getattr(model_sharded_by, 'django_sharding__shard_field')
        return getattr(model_sharded_by, sharded_field)


class RoundRobinBucketingStrategy(BaseShardedModelBucketingStrategy):
    """
    A shard selection strategy that assigns shards in a round-robin way.
    This is non-deterministic and this strategy assumes the shard is saved to
    the model.
    """
    def __init__(self, shard_group, databases):
        super(RoundRobinBucketingStrategy, self).__init__(shard_group)
        shards = self.get_shards(databases)
        # To help balance out the fact that this would start the round-robin
        # bucketing at initialization time, we start at a random index.
        max_index = max(0, len(shards) - 1)
        starting_index = randint(0, max_index)
        shards = shards[starting_index:] + shards[:starting_index]
        self._shards_cycle = cycle(shards)

    def pick_shard(self, model_sharded_by):
        return next(self._shards_cycle)


class RandomBucketingStrategy(BaseShardedModelBucketingStrategy):
    """
    A shard selection strategy that assigns shards randomly.
    This is non-deterministic and this strategy assumes the shard is saved to
    the model.
    """
    def __init__(self, shard_group, databases):
        super(RandomBucketingStrategy, self).__init__(shard_group)
        self.shards = self.get_shards(databases)

    def pick_shard(self, model_sharded_by):
        return choice(self.shards)


class ModBucketingStrategy(BaseBucketingStrategy):
    """
    A shard selection strategy that assigns shards based on the mod of the
    models pk.
    Note: It is only deterministic as long as the number of shards do not
    change.
    """
    def __init__(self, shard_group, databases):
        super(ModBucketingStrategy, self).__init__(shard_group)
        self.shards = self.get_shards(databases)

    def pick_shard(self, model_sharded_by):
        return self.shards[hash(str(model_sharded_by.pk)) % len(self.shards)]

    def get_shard(self, model_sharded_by):
        return self.pick_shard(model_sharded_by)


class SavedModBucketingStrategy(BaseShardedModelBucketingStrategy, ModBucketingStrategy):
    """
    A shard selection strategy that assigns shards based on the mod of the
    models pk.
    This is non-deterministic and this strategy assumes the shard is saved to
    the model. It is non-deterministic as the number of shards may change.
    """
    pass



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-07 23:12
from __future__ import unicode_literals

import django.contrib.auth.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='user',
            name='username',
            field=models.CharField(error_messages={'unique': 'A user with that username already exists.'}, help_text='Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.', max_length=150, unique=True, validators=[django.contrib.auth.validators.ASCIIUsernameValidator()], verbose_name='username'),
        ),
    ]



from dj_database_url import config

from django.test import TestCase

from django_sharding_library.settings_helpers import database_config, database_configs


class DatabaseConfigTestCase(TestCase):

    def setUp(self):
        self.default_database_url = 'postgres://user:pass@localhost/test_db'
        self.dj_database_config = config(env='TEST_ENV_VAR', default=self.default_database_url)

    def test_returns_none_when_no_env_var_and_no_url(self):
        result = database_config(environment_variable='TEST_ENV_VAR', default_database_url=None, shard_group='default')
        expected_result = {}
        self.assertEqual(result, expected_result)

    def test_no_extra_args_equivalent_to_dj_database_with_non_shard_status_added(self):
        result = database_config(environment_variable='TEST_ENV_VAR', default_database_url=self.default_database_url)
        expected_result = {'SHARD_GROUP': None, 'TEST': {}}
        expected_result.update(self.dj_database_config)
        self.assertEqual(result, expected_result)

    def test_adds_sharded_state_to_dict_when_sharded(self):
        result = database_config(environment_variable='TEST_ENV_VAR', default_database_url=self.default_database_url, shard_group='default')
        expected_result = {'SHARD_GROUP': 'default', 'TEST': {}}
        expected_result.update(self.dj_database_config)
        self.assertEqual(result, expected_result)

    def test_adds_replica_state_to_dict(self):
        result = database_config(environment_variable='TEST_ENV_VAR', default_database_url=self.default_database_url, is_replica_of='some_database')
        expected_result = {'PRIMARY': 'some_database', 'SHARD_GROUP': None, 'TEST': {'MIRROR': 'some_database'}}
        expected_result.update(self.dj_database_config)
        self.assertEqual(result, expected_result)

    def test_adds_replica_state_to_dict_also_when_sharded(self):
        result = database_config(environment_variable='TEST_ENV_VAR', default_database_url=self.default_database_url, shard_group='default_set_of_shards', is_replica_of='some_database')
        expected_result = {'PRIMARY': 'some_database', 'SHARD_GROUP': 'default_set_of_shards', 'TEST': {'MIRROR': 'some_database'}}
        expected_result.update(self.dj_database_config)
        self.assertEqual(result, expected_result)


class DatabaseConfigsTestCase(TestCase):

    def setUp(self):
        self.default_database_url = 'postgres://user:pass@localhost/test_db'
        self.dj_database_config = config(env='TEST_ENV_VAR', default=self.default_database_url)

    def test_skips_databases_with_no_envvar_value_or_deafult_url(self):

        simple_config = {
            'unsharded_databases': [
                {
                    'name': 'DB01',
                    'environment_variable': 'ENV',
                    'default_database_url': ''
                }
            ]
        }
        result = database_configs(simple_config)
        expected_result = {}
        self.assertEqual(result, expected_result)

    def test_databases_with_envvar_value_and_no_default_url(self):
        import os
        os.environ['SOME_USELESS_ENV'] = self.default_database_url

        simple_config = {
            'unsharded_databases': [
                {
                    'name': 'DB01',
                    'environment_variable': 'SOME_USELESS_ENV',
                    'default_database_url': ''
                }
            ]
        }
        result = database_configs(simple_config)
        del os.environ['SOME_USELESS_ENV']

        DB01 = {'SHARD_GROUP': None, 'TEST': {}}
        DB01.update(self.dj_database_config)

        self.assertEqual(result, {'DB01': DB01})

    def test_databases_with_default_url_unset_env(self):
        simple_config = {
            'unsharded_databases': [
                {
                    'name': 'DB01',
                    'environment_variable': 'SOME_OTHER_USELESS_ENV',
                    'default_database_url': self.default_database_url
                }
            ]
        }
        result = database_configs(simple_config)

        DB01 = {'SHARD_GROUP': None, 'TEST': {}}
        DB01.update(self.dj_database_config)

        self.assertEqual(result, {'DB01': DB01})

    def test_unsharded_databases_ignore_shard_group(self):
        simple_config = {
            'unsharded_databases': [
                {
                    'name': 'DB01',
                    'environment_variable': 'SOME_OTHER_USELESS_ENV',
                    'default_database_url': self.default_database_url,
                    'shard_group': 'testing'
                }
            ]
        }
        result = database_configs(simple_config)

        DB01 = {'SHARD_GROUP': None, 'TEST': {}}
        DB01.update(self.dj_database_config)

        self.assertEqual(result, {'DB01': DB01})

    def test_sharded_databases_default_shard_group(self):
        simple_config = {
            'sharded_databases': [
                {
                    'name': 'DB01',
                    'environment_variable': 'SOME_OTHER_USELESS_ENV',
                    'default_database_url': self.default_database_url,
                }
            ]
        }
        result = database_configs(simple_config)

        DB01 = {'SHARD_GROUP': 'default', 'TEST': {}, 'SHARD_ID': 0}
        DB01.update(self.dj_database_config)

        self.assertEqual(result, {'DB01': DB01})

    def test_sharded_databases_shard_group(self):
        simple_config = {
            'sharded_databases': [
                {
                    'name': 'DB01',
                    'environment_variable': 'SOME_OTHER_USELESS_ENV',
                    'default_database_url': self.default_database_url,
                    'shard_group': 'testing'
                }
            ]
        }
        result = database_configs(simple_config)

        DB01 = {'SHARD_GROUP': 'testing', 'TEST': {}, 'SHARD_ID': 0}
        DB01.update(self.dj_database_config)

        self.assertEqual(result, {'DB01': DB01})

    def test_sharded_databases_shard_id(self):
        simple_config = {
            'sharded_databases': [
                {
                    'name': 'DB01',
                    'environment_variable': 'SOME_OTHER_USELESS_ENV',
                    'default_database_url': self.default_database_url,
                    'shard_group': 'testing'
                },
                {
                    'name': 'DB02',
                    'environment_variable': 'SOME_OTHER_USELESS_ENV',
                    'default_database_url': self.default_database_url,
                    'shard_group': 'testing'
                },
                {
                    'name': 'DB03',
                    'environment_variable': 'SOME_OTHER_USELESS_ENV',
                    'default_database_url': self.default_database_url,
                    'shard_group': 'default'
                },
                {
                    'name': 'DB04',
                    'environment_variable': 'SOME_OTHER_USELESS_ENV',
                    'default_database_url': self.default_database_url,
                    'shard_group': 'default'
                },
                {
                    'name': 'DB05',
                    'environment_variable': 'SOME_OTHER_USELESS_ENV',
                    'default_database_url': self.default_database_url,
                    'shard_group': 'default'
                }
            ]
        }
        result = database_configs(simple_config)

        DB01 = {'SHARD_GROUP': 'testing', 'TEST': {}, 'SHARD_ID': 0}
        DB02 = {'SHARD_GROUP': 'testing', 'TEST': {}, 'SHARD_ID': 1}
        DB03 = {'SHARD_GROUP': 'default', 'TEST': {}, 'SHARD_ID': 0}
        DB04 = {'SHARD_GROUP': 'default', 'TEST': {}, 'SHARD_ID': 1}
        DB05 = {'SHARD_GROUP': 'default', 'TEST': {}, 'SHARD_ID': 2}
        DB01.update(self.dj_database_config)
        DB02.update(self.dj_database_config)
        DB03.update(self.dj_database_config)
        DB04.update(self.dj_database_config)
        DB05.update(self.dj_database_config)

        self.assertEqual(result['DB01'], DB01)
        self.assertEqual(result['DB02'], DB02)
        self.assertEqual(result['DB03'], DB03)
        self.assertEqual(result['DB04'], DB04)
        self.assertEqual(result['DB05'], DB05)

    def test_unsharded_replica_database(self):
        simple_config = {
            'unsharded_databases': [
                {
                    'name': 'DB01',
                    'environment_variable': 'SOME_OTHER_USELESS_ENV',
                    'default_database_url': self.default_database_url,
                    'replicas': [{
                        'name': 'DB01_replica',
                        'environment_variable': 'SOME_OTHER_USELESS_ENV',
                        'default_database_url': self.default_database_url,
                    }]
                }
            ]
        }
        result = database_configs(simple_config)

        DB01 = {'SHARD_GROUP': None, 'TEST': {}}
        DB01.update(self.dj_database_config)
        DB01_replica = {'SHARD_GROUP': None, 'PRIMARY': 'DB01', 'TEST': {'MIRROR': 'DB01'}}
        DB01_replica.update(self.dj_database_config)

        self.assertEqual(result, {'DB01': DB01, 'DB01_replica': DB01_replica})

    def test_sharded_replica_database(self):
        simple_config = {
            'sharded_databases': [
                {
                    'name': 'DB01',
                    'environment_variable': 'SOME_OTHER_USELESS_ENV',
                    'default_database_url': self.default_database_url,
                    'shard_group': 'testing',
                    'replicas': [{
                        'name': 'DB01_replica',
                        'environment_variable': 'SOME_OTHER_USELESS_ENV',
                        'default_database_url': self.default_database_url,
                    }]
                }
            ]
        }
        result = database_configs(simple_config)

        DB01 = {'SHARD_GROUP': 'testing', 'TEST': {}, 'SHARD_ID': 0}
        DB01.update(self.dj_database_config)
        DB01_replica = {'SHARD_GROUP': 'testing', 'PRIMARY': 'DB01', 'TEST': {'MIRROR': 'DB01'}}
        DB01_replica.update(self.dj_database_config)

        self.assertEqual(result, {'DB01': DB01, 'DB01_replica': DB01_replica})



import os

from django.conf import settings
from django.test import TransactionTestCase

from django_sharding_library.constants import Backends


class TravisTestCase(TransactionTestCase):

    def test_travis_uses_non_sqlite_databases(self):
        TRAVISCI = os.environ.get('TRAVIS')

        self.assertEqual(len(settings.DATABASES), 7)

        if TRAVISCI:
            self.assertIn(settings.DATABASES['default']['ENGINE'], Backends.POSTGRES)
            self.assertIn(settings.DATABASES['app_shard_001']['ENGINE'], Backends.MYSQL)
            self.assertIn(settings.DATABASES['app_shard_001_replica_001']['ENGINE'], Backends.POSTGRES)
            self.assertIn(settings.DATABASES['app_shard_001_replica_002']['ENGINE'], Backends.POSTGRES)
            self.assertIn(settings.DATABASES['app_shard_002']['ENGINE'], Backends.POSTGRES)
            self.assertIn(settings.DATABASES['app_shard_003']['ENGINE'], Backends.POSTGRES)
            self.assertIn(settings.DATABASES['app_shard_004']['ENGINE'], Backends.POSTGRES)
        else:
            self.assertIn(settings.DATABASES['default']['ENGINE'], Backends.SQLITE)
            self.assertIn(settings.DATABASES['app_shard_001']['ENGINE'], Backends.SQLITE)
            self.assertIn(settings.DATABASES['app_shard_001_replica_001']['ENGINE'], Backends.SQLITE)
            self.assertIn(settings.DATABASES['app_shard_001_replica_002']['ENGINE'], Backends.SQLITE)
            self.assertIn(settings.DATABASES['app_shard_002']['ENGINE'], Backends.SQLITE)
            self.assertIn(settings.DATABASES['app_shard_003']['ENGINE'], Backends.SQLITE)
            self.assertIn(settings.DATABASES['app_shard_004']['ENGINE'], Backends.SQLITE)



class DjangoShardingException(Exception):
    pass


class ShardedModelInitializationException(DjangoShardingException):
    pass


class InvalidMigrationException(DjangoShardingException):
    pass


class NonExistentDatabaseException(DjangoShardingException):
    pass



from setuptools import setup, find_packages

version = '0.1.0'


setup(
    name='django_sharding',
    version=version,
    description="""A sharding app to include in your project so that you can shard your data.""",
    long_description=open('README.md').read(),
    author='JBKahn',
    author_email='josephbkahn@gmail.com',
    url='https://github.com/JBKahn/django-sharding',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['Django>=1.8', 'dj-database-url==0.4.1'],
    tests_require=['psycopg2==2.6.2', 'mysqlclient==1.3.7', 'mock==2.0.0', 'django_nose==1.4.4', 'tox==2.3.1'],
    license="BSD",
    zip_safe=False,
    keywords='django shard sharding library',
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'
    ],
    test_suite='runtests.run_tests',
)



from django.db import connections, DatabaseError, transaction
from django.conf import settings
from django_sharding_library.sql import postgres_shard_id_function_sql
from django.db.models import signals

from django_sharding_library.exceptions import DjangoShardingException



def create_postgres_global_sequence(sequence_name, db_alias, reset_sequence=False):
    cursor = connections[db_alias].cursor()
    sid = transaction.savepoint(db_alias)
    create_sequence_if_not_exists_sql = """DO
$$
BEGIN
        CREATE SEQUENCE %s;
EXCEPTION WHEN duplicate_table THEN
        -- do nothing, it's already there
END
$$ LANGUAGE plpgsql;"""
    try:
        cursor.execute(create_sequence_if_not_exists_sql % sequence_name)
    except DatabaseError:
        transaction.savepoint_rollback(sid, using=db_alias)
    else:
        transaction.savepoint_commit(sid, using=db_alias)
    if reset_sequence:
        cursor.execute("SELECT setval('%s', 1, false)" % (sequence_name,))
    cursor.close()


def create_postgres_shard_id_function(sequence_name, db_alias, shard_id):
    cursor = connections[db_alias].cursor()
    cursor.execute(postgres_shard_id_function_sql % {'shard_epoch': settings.SHARD_EPOCH,
                                                     'shard_id': shard_id,
                                                     'sequence_name': sequence_name})
    cursor.close()


def verify_postres_id_field_setup_correctly(sequence_name, db_alias, function_name):
    cursor = connections[db_alias].cursor()
    cursor.execute(
        "SELECT count(*) FROM pg_class c WHERE c.relkind = '%s' and c.relname = '%s';" % ('S', sequence_name)
    )

    if cursor.fetchone()[0] == 0:
        cursor.close()
        return False

    cursor.execute(
        "SELECT count(*) from pg_proc p where p.proname = '%s';" % (function_name,)
    )

    if cursor.fetchone()[0] == 0:
        cursor.close()
        return False

    cursor.close()
    return True


def register_migration_signal_for_model_receiver(model, function, dispatch_uid=None):
    signals.pre_migrate.connect(function, sender=model, dispatch_uid=dispatch_uid)


def is_model_class_on_database(model, database):
    specific_database = getattr(model, 'django_sharding__database', None)
    is_sharded = getattr(model, 'django_sharding__is_sharded', False)

    if specific_database and is_sharded:
        raise DjangoShardingException('Model marked as both sharded and on a single database, unable to determine where to run migrations for {}.'.format(model.__class__.__name__))

    if specific_database:
        return getattr(model, 'django_sharding__database') == database

    if is_sharded:
        shard_group = getattr(model, 'django_sharding__shard_group', None)
        if shard_group:
            return settings.DATABASES[database]['SHARD_GROUP'] == shard_group
        raise DjangoShardingException("Unable to determine what database the model is on as the shard group of {} is unknown.".format(model.__class__.__name__))

    return database == "default"


def get_possible_databases_for_model(model):
    return [
        database for database in settings.DATABASES
        if is_model_class_on_database(model=model, database=database)
    ]


def get_database_for_model_instance(instance):
    if instance._state.db:
        return instance._state.db

    model = instance._meta.model
    possible_databases = get_possible_databases_for_model(model=model)
    if len(possible_databases) == 1:
        return possible_databases[0]
    elif len(possible_databases) == 0:
        pass
    else:
        model_has_sharded_id_field = getattr(model, 'django_sharding__sharded_by_field', None) is not None

        if model_has_sharded_id_field:
            sharded_by_field_id = getattr(instance, getattr(model, 'django_sharding__sharded_by_field', 'django_sharding__none'), None)
            if sharded_by_field_id is not None:
                return model.get_shard_from_id(sharded_by_field_id)

        return instance.get_shard()

    raise DjangoShardingException("Unable to deduce datbase for model instance")



from mock import call, patch
import unittest

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.test import TransactionTestCase

from tests.models import TestModel, ShardedTestModelIDs, PostgresCustomIDModel, PostgresShardUser
from django_sharding_library.exceptions import InvalidMigrationException
from django_sharding_library.router import ShardedRouter
from django_sharding_library.routing_read_strategies import BaseRoutingStrategy
from django_sharding_library.fields import PostgresShardGeneratedIDField
from django_sharding_library.constants import Backends
from django_sharding_library.manager import ShardManager

from django.db.models import Sum, Q


class FakeRoutingStrategy(BaseRoutingStrategy):
    """
    A fake router for testing that will always return
    the `testing` DB rather than an existing database.
    """
    def pick_read_db(self, primary_db_name):
        return 'testing'


class RouterReadTestCase(TransactionTestCase):

    def setUp(self):
        from django.contrib.auth import get_user_model
        self.sut = ShardedRouter()
        self.user = get_user_model().objects.create_user(username='username', password='pwassword', email='test@example.com')

    def test_sharded_instance_uses_a_routing_read_strategies(self):
        item = TestModel.objects.using('app_shard_002').create(random_string=2, user_pk=self.user.pk)
        hints = {'instance': item}
        with patch.object(self.sut, 'get_read_db_routing_strategy') as mock_get_read_db_routing_strategy:
            mock_get_read_db_routing_strategy.return_value = FakeRoutingStrategy(settings.DATABASES)
            self.assertEqual(self.sut.db_for_read(model=TestModel, **hints), 'testing')

    def test_sharded_instance(self):
        item = TestModel.objects.using('app_shard_002').create(random_string=2, user_pk=self.user.pk)
        hints = {'instance': item}
        self.assertEqual(self.sut.db_for_read(model=TestModel, **hints), 'app_shard_002')

    def test_sharded_model(self):
        self.assertEqual(self.sut.db_for_read(model=TestModel), None)

    def test_specific_database(self):
        self.assertEqual(self.sut.db_for_read(model=ShardedTestModelIDs), 'app_shard_001')

    def test_specific_database_does_not_use_the_routing_read_strategy(self):
        with patch.object(self.sut, 'get_read_db_routing_strategy') as mock_get_read_db_routing_strategy:
            mock_get_read_db_routing_strategy.return_value = FakeRoutingStrategy(settings.DATABASES)
            self.assertEqual(self.sut.db_for_read(model=ShardedTestModelIDs), 'app_shard_001')

    def test_other(self):
        from django.contrib.auth import get_user_model
        self.assertEqual(self.sut.db_for_read(model=get_user_model()), "default")

    def test_router_hints_receives_get_kwargs(self):
        original_id = TestModel.objects.create(user_pk=self.user.pk).id

        lookups_to_find = {'exact_lookups': {'user_pk': self.user.pk}}

        with patch.object(ShardedRouter, 'db_for_write', wraps=self.sut.db_for_write) as write_route_function:
            with patch.object(ShardedRouter, 'db_for_read', wraps=self.sut.db_for_read) as read_route_function:
                result = TestModel.objects.get(user_pk=self.user.pk)

        self.assertEqual(result.id, original_id)
        self.assertEqual(
            [call(TestModel, **lookups_to_find), call(get_user_model())],
            read_route_function.mock_calls
        )
        self.assertEqual(
            [],
            write_route_function.mock_calls
        )

    def test_router_hints_receives_get_kwargs_on_get_or_create__get(self):
        original_id = TestModel.objects.create(user_pk=self.user.pk).id

        lookups_to_find = {'exact_lookups': {'user_pk': self.user.pk}}

        with patch.object(ShardedRouter, 'db_for_write', wraps=self.sut.db_for_write) as write_route_function:
            with patch.object(ShardedRouter, 'db_for_read', wraps=self.sut.db_for_read) as read_route_function:
                result, created = TestModel.objects.get_or_create(user_pk=self.user.pk)

        self.assertEqual(result.id, original_id)
        self.assertFalse(created)
        self.assertEqual(
            [call(get_user_model())],
            read_route_function.mock_calls
        )

        self.assertEqual(
            [
                call(TestModel, **lookups_to_find),
            ],
            write_route_function.mock_calls
        )

    def test_router_hints_receives_get_kwargs_on_get_or_create__create(self):
        lookups_to_find = {'exact_lookups': {'user_pk': self.user.pk}}

        with patch.object(ShardedRouter, 'db_for_write', wraps=self.sut.db_for_write) as write_route_function:
            with patch.object(ShardedRouter, 'db_for_read', wraps=self.sut.db_for_read) as read_route_function:

                _, created = TestModel.objects.get_or_create(user_pk=self.user.pk)

        self.assertTrue(created)
        self.assertEqual(
            [call(get_user_model()), call(get_user_model()), call(get_user_model())],
            read_route_function.mock_calls
        )

        self.assertEqual(
            [
                call(TestModel, **lookups_to_find),
                call(TestModel, **lookups_to_find),
                call(TestModel, instance=write_route_function.mock_calls[2][2]["instance"], **lookups_to_find),  # no way to access that copy of the instance here, the one prior to saving.
                call(ShardedTestModelIDs),
            ],
            write_route_function.mock_calls
        )

    def test_router_hints_receives_get_kwargs_on_update_or_create__get(self):
        original_id = TestModel.objects.create(user_pk=self.user.pk).id

        lookups_to_find = {'exact_lookups': {'user_pk': self.user.pk}}

        with patch.object(ShardedRouter, 'db_for_read', wraps=self.sut.db_for_read) as read_route_function:
            with patch.object(ShardedRouter, 'db_for_write', wraps=self.sut.db_for_write) as write_route_function:

                result, created = TestModel.objects.update_or_create(user_pk=self.user.pk)

        self.assertEqual(result.id, original_id)
        self.assertFalse(created)

        # Django 1.10 dropped a call for each here :)
        import django
        if django.VERSION < (1, 10):
            self.assertEqual(
                [call(get_user_model()), call(get_user_model()), call(get_user_model())],
                read_route_function.mock_calls
            )
            self.assertEqual(
                [call(TestModel, **lookups_to_find), call(TestModel, **lookups_to_find), call(TestModel, **lookups_to_find)],
                write_route_function.mock_calls
            )

        else:
            self.assertEqual(
                [call(get_user_model()), call(get_user_model())],
                read_route_function.mock_calls
            )

            self.assertEqual(
                [call(TestModel, **lookups_to_find), call(TestModel, **lookups_to_find)],
                write_route_function.mock_calls
            )

    def test_router_hints_receives_get_kwargs_on_update_or_create__create(self):
        lookups_to_find = {'exact_lookups': {'user_pk': self.user.pk}}

        with patch.object(ShardedRouter, 'db_for_read', wraps=self.sut.db_for_read) as read_route_function:
            with patch.object(ShardedRouter, 'db_for_write', wraps=self.sut.db_for_write) as write_route_function:

                instance, created = TestModel.objects.update_or_create(user_pk=self.user.pk)

        self.assertTrue(created)
        self.assertEqual(
            [call(get_user_model()), call(get_user_model()), call(get_user_model())],
            read_route_function.mock_calls
        )

        self.assertEqual(
            [
                call(TestModel, **lookups_to_find),
                call(TestModel, **lookups_to_find),
                call(TestModel, instance=write_route_function.mock_calls[2][2]["instance"], **lookups_to_find),  # no way to access that copy of the instance here, the one prior to saving.
                call(ShardedTestModelIDs),
            ],
            write_route_function.mock_calls
        )

    def test_router_hints_receives_filter_kwargs_on_count(self):
        TestModel.objects.create(user_pk=self.user.pk)

        lookups_to_find = {'exact_lookups': {'user_pk': self.user.pk}}

        with patch.object(ShardedRouter, 'db_for_write', wraps=self.sut.db_for_write) as write_route_function:
            with patch.object(ShardedRouter, 'db_for_read', wraps=self.sut.db_for_read) as read_route_function:
                result = TestModel.objects.filter(user_pk=self.user.pk).count()

        self.assertEqual(result, 1)
        self.assertEqual(
            [call(TestModel, **lookups_to_find), call(get_user_model())],
            read_route_function.mock_calls
        )
        self.assertEqual(
            [],
            write_route_function.mock_calls
        )

    def test_router_hints_receives_filter_kwargs_on_exists(self):
        TestModel.objects.create(user_pk=self.user.pk)

        lookups_to_find = {'exact_lookups': {'user_pk': self.user.pk}}

        with patch.object(ShardedRouter, 'db_for_write', wraps=self.sut.db_for_write) as write_route_function:
            with patch.object(ShardedRouter, 'db_for_read', wraps=self.sut.db_for_read) as read_route_function:
                result = TestModel.objects.filter(user_pk=self.user.pk).exists()

        self.assertTrue(result)
        self.assertEqual(
            [call(TestModel, **lookups_to_find), call(get_user_model())],
            read_route_function.mock_calls
        )
        self.assertEqual(
            [],
            write_route_function.mock_calls
        )

    def test_router_hints_receives_filter_kwargs(self):
        TestModel.objects.create(user_pk=self.user.pk)

        lookups_to_find = {'exact_lookups': {'user_pk': self.user.pk}}

        with patch.object(ShardedRouter, 'db_for_write', wraps=self.sut.db_for_write) as write_route_function:
            with patch.object(ShardedRouter, 'db_for_read', wraps=self.sut.db_for_read) as read_route_function:
                list(TestModel.objects.filter(user_pk=self.user.pk))

        self.assertEqual(
            [call(TestModel, **lookups_to_find), call(get_user_model())],
            read_route_function.mock_calls
        )
        self.assertEqual(
            [],
            write_route_function.mock_calls
        )

    def test_router_gets_hints_correctly_with_positional_arguments_like_Q_in_filter(self):
        TestModel.objects.create(user_pk=self.user.pk, random_string="test")

        lookups_to_find = {'exact_lookups': {'user_pk': self.user.pk}}

        with patch.object(ShardedRouter, 'db_for_write', wraps=self.sut.db_for_write) as write_route_function:
            with patch.object(ShardedRouter, 'db_for_read', wraps=self.sut.db_for_read) as read_route_function:
                list(TestModel.objects.filter(Q(random_string="test") | Q(random_string__isnull=True), user_pk=self.user.pk))

        self.assertEqual(
            [call(TestModel, **lookups_to_find), call(get_user_model())],
            read_route_function.mock_calls
        )
        self.assertEqual(
            [],
            write_route_function.mock_calls
        )

    def test_router_gets_hints_correctly_with_positional_arguments_like_Q_in_get(self):
        TestModel.objects.create(user_pk=self.user.pk, random_string="test")

        lookups_to_find = {'exact_lookups': {'user_pk': self.user.pk}}

        with patch.object(ShardedRouter, 'db_for_write', wraps=self.sut.db_for_write) as write_route_function:
            with patch.object(ShardedRouter, 'db_for_read', wraps=self.sut.db_for_read) as read_route_function:
                TestModel.objects.get(Q(random_string="test") | Q(random_string__isnull=True), user_pk=self.user.pk)

        self.assertEqual(
            [call(TestModel, **lookups_to_find), call(get_user_model())],
            read_route_function.mock_calls
        )
        self.assertEqual(
            [],
            write_route_function.mock_calls
        )

    def test_queryset_router_filter_returns_existing_objects(self):
        for i in range(1, 11):
            test_model_obj = TestModel.objects.create(user_pk=self.user.pk, random_string="%s" % i)
            self.assertIn(test_model_obj._state.db, ['app_shard_001', 'app_shard_002'])

        test_models = TestModel.objects.filter(user_pk=self.user.pk)
        self.assertEqual(len(test_models), 10)

        from django.contrib.auth import get_user_model
        new_user = get_user_model().objects.create_user(username='username2', password='pwassword', email='test2@example.com')
        for i in range(1, 21):
            TestModel.objects.create(user_pk=new_user.pk, random_string="%s" % i)

        test_models = TestModel.objects.filter(user_pk=new_user.pk)
        self.assertEqual(len(test_models), 20)

    def test_queryset_router_filter_with_aggregates(self):
        for i in range(1, 11):
            TestModel.objects.create(user_pk=self.user.pk, random_string="%s" % i)
        num_models = TestModel.objects.filter(user_pk=self.user.pk).count()
        self.assertEqual(num_models, 10)

        sum_model_pk = TestModel.objects.filter(user_pk=self.user.pk).aggregate(user_pk_sum=Sum('user_pk'))
        self.assertEqual(sum_model_pk['user_pk_sum'], self.user.pk * 10)


class RouterWriteTestCase(TransactionTestCase):

    def setUp(self):
        from django.contrib.auth import get_user_model
        self.sut = ShardedRouter()
        self.user = get_user_model().objects.create_user(username='username', password='pwassword', email='test@example.com')

    def test_sharded_instance_does_not_use_a_routing_read_strategies(self):
        item = TestModel.objects.using('app_shard_002').create(random_string=2, user_pk=self.user.pk)
        hints = {'instance': item}
        with patch.object(self.sut, 'get_read_db_routing_strategy') as mock_get_read_db_routing_strategy:
            mock_get_read_db_routing_strategy.return_value = FakeRoutingStrategy(settings.DATABASES)
            self.assertEqual(self.sut.db_for_write(model=TestModel, **hints), 'app_shard_002')

    def test_sharded_instance(self):
        item = TestModel.objects.using('app_shard_002').create(random_string=2, user_pk=self.user.pk)
        hints = {'instance': item}
        self.assertEqual(self.sut.db_for_write(model=TestModel, **hints), 'app_shard_002')

    def test_sharded_model(self):
        self.assertEqual(self.sut.db_for_write(model=TestModel), None)

    def test_specific_database(self):
        self.assertEqual(self.sut.db_for_write(model=ShardedTestModelIDs), 'app_shard_001')

    def test_specific_database_does_not_use_the_routing_read_strategy(self):
        with patch.object(self.sut, 'get_read_db_routing_strategy') as mock_get_read_db_routing_strategy:
            mock_get_read_db_routing_strategy.return_value = FakeRoutingStrategy(settings.DATABASES)
            self.assertEqual(self.sut.db_for_write(model=ShardedTestModelIDs), 'app_shard_001')

    def test_other(self):
        from django.contrib.auth import get_user_model
        self.assertEqual(self.sut.db_for_write(model=get_user_model()), "default")

    def test_create_sharded_object_without_using(self):
        instance = TestModel.objects.create(user_pk=self.user.pk)
        self.assertEqual(instance._state.db, self.user.shard)
        self.assertTrue(TestModel.objects.using(self.user.shard).get(id=instance.id))


class RouterAllowRelationTestCase(TransactionTestCase):

    def setUp(self):
        from django.contrib.auth import get_user_model
        self.sut = ShardedRouter()
        self.user = get_user_model().objects.create_user(username='username', password='pwassword', email='test@example.com')

    def test_allow_relation_two_items_on_the_same_non_default_database(self):
        item_one = ShardedTestModelIDs.objects.create(stub=None)
        item_two = ShardedTestModelIDs.objects.create(stub=None)
        self.assertTrue(self.sut.allow_relation(item_one, item_two))

    def test_do_not_allow_relation_of_item_on_specific_database__and_non_sharded_non_specific_database_instances(self):
        item_one = ShardedTestModelIDs.objects.create(stub=None)
        self.assertFalse(self.sut.allow_relation(item_one, self.user))

    def test_do_not_allow_relation_of_sharded_models_on_different_shards(self):
        item_one = TestModel.objects.using('app_shard_002').create(random_string=2, user_pk=self.user.pk)
        item_two = TestModel.objects.using('app_shard_001').create(random_string=2, user_pk=self.user.pk)
        self.assertFalse(self.sut.allow_relation(item_one, item_two))

    def test_allow_relation_of_sharded_models_on_same_shard(self):
        item_one = TestModel.objects.using('app_shard_001').create(random_string=2, user_pk=self.user.pk)
        item_two = TestModel.objects.using('app_shard_001').create(random_string=2, user_pk=self.user.pk)
        self.assertTrue(self.sut.allow_relation(item_one, item_two))

    def test_allow_relation_of_sharded_instance_and_item_on_specific_db(self):
        item_one = TestModel.objects.using('app_shard_001').create(random_string=2, user_pk=self.user.pk)
        item_two = ShardedTestModelIDs.objects.create(stub=None)
        self.assertTrue(self.sut.allow_relation(item_one, item_two))

    def test_do_not_allow_relation_of_sharded_instance_and_item_on_specific_db_when_on_different_dbs(self):
        item_one = TestModel.objects.using('app_shard_002').create(random_string=2, user_pk=self.user.pk)
        item_two = ShardedTestModelIDs.objects.create(stub=None)
        self.assertFalse(self.sut.allow_relation(item_one, item_two))

    def test_do_not_allow_relation_of_sharded_instance_and_non_sharded_non_specific_database_instances(self):
        item_one = TestModel.objects.using('app_shard_001').create(random_string=2, user_pk=self.user.pk)
        self.assertFalse(self.sut.allow_relation(item_one, self.user))

    def test_allow_relation_of_items_on_default_database(self):
        item_one = Group.objects.create(name='test_group')
        self.assertTrue(self.sut.allow_relation(item_one, self.user))


class RouterAllowMigrateTestCase(TransactionTestCase):

    def setUp(self):
        self.sut = ShardedRouter()

    def assert_allow_migrate(self, app_label, model_name, can_migrate_default, can_migrate_shard, migratable_db=None):
        self.assertEqual(self.sut.allow_migrate(db='default', app_label=app_label, model_name=model_name), can_migrate_default)
        self.assertEqual(
            all([
                self.sut.allow_migrate(db='app_shard_001_replica_001', app_label=app_label, model_name=model_name),
                self.sut.allow_migrate(db='app_shard_001_replica_002', app_label=app_label, model_name=model_name),
            ]),
            False
        )
        self.assertEqual(
            all([
                self.sut.allow_migrate(db='app_shard_001', app_label=app_label, model_name=model_name),
                self.sut.allow_migrate(db='app_shard_002', app_label=app_label, model_name=model_name),
            ]),
            can_migrate_shard
        )
        if migratable_db:
            self.assertTrue(self.sut.allow_migrate(db=migratable_db, app_label=app_label, model_name=model_name))

    def test_model_name_passed_in(self):
        self.assertTrue(self.sut.allow_migrate(db='default', app_label='tests', model_name="User"))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001', app_label='tests', model_name="User"))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_002', app_label='tests', model_name="User"))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001_replica_001', app_label='tests', model_name="User"))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001_replica_002', app_label='tests', model_name="User"))

    def test_model_passed_in(self):
        from django.contrib.auth import get_user_model

        hints = {'model': get_user_model()}

        self.assertTrue(self.sut.allow_migrate(db='default', app_label='tests', model_name=None, **hints))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001', app_label='tests', model_name=None, **hints))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_002', app_label='tests', model_name=None, **hints))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001_replica_001', app_label='tests', model_name=None, **hints))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001_replica_002', app_label='tests', model_name=None, **hints))

    def test_app_passed_in(self):
        self.assertTrue(self.sut.allow_migrate(db='default', app_label='tests'))
        self.assertTrue(self.sut.allow_migrate(db='app_shard_001', app_label='tests'))
        self.assertTrue(self.sut.allow_migrate(db='app_shard_002', app_label='tests'))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001_replica_001', app_label='tests'))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001_replica_002', app_label='tests'))

    def test_force_migrate_on_databases(self):
        hints = {"force_migrate_on_databases": ["app_shard_001"]}
        self.assertFalse(self.sut.allow_migrate(db='default', app_label='tests', **hints))
        self.assertTrue(self.sut.allow_migrate(db='app_shard_001', app_label='tests', **hints))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_002', app_label='tests', **hints))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001_replica_001', app_label='tests', **hints))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001_replica_002', app_label='tests', **hints))

    def test_force_migrate_on_databases_ignores_secondary(self):
        hints = {"force_migrate_on_databases": ["app_shard_001_replica_001"]}
        self.assertFalse(self.sut.allow_migrate(db='default', app_label='tests', **hints))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001', app_label='tests', **hints))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_002', app_label='tests', **hints))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001_replica_001', app_label='tests', **hints))
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001_replica_002', app_label='tests', **hints))

    def test_migrate_replica_will_not_work(self):
        self.assertFalse(self.sut.allow_migrate(db='app_shard_001_replica_001', app_label='tests', model_name='TestModel'))
        self.assertTrue(self.sut.allow_migrate(db='app_shard_001', app_label='tests', model_name='TestModel'))

    def test_migrate_sharded_model_with_specific_database_will_not_work(self):
        sut = ShardedRouter()

        with patch.object(TestModel, "django_sharding__database", 'blah', create=True):
            with patch.object(TestModel, "django_sharding__is_sharded", True, create=True):
                with self.assertRaises(InvalidMigrationException):
                    sut.allow_migrate(db='default', app_label='tests', model_name='TestModel')

        with patch.object(TestModel, "django_sharding__database", 'blah', create=True):
            with patch.object(TestModel, "django_sharding__is_sharded", False, create=True):
                sut.allow_migrate(db='default', app_label='tests', model_name='TestModel')

        with patch.object(TestModel, "django_sharding__database", None, create=True):
            with patch.object(TestModel, "django_sharding__is_sharded", True, create=True):
                sut.allow_migrate(db='default', app_label='tests', model_name='TestModel')

        sut.allow_migrate(db='default', app_label='tests', model_name='TestModel')

    def test_django_model_only_allows_on_default(self):
        self.assert_allow_migrate(
            app_label='auth',
            model_name='Group',
            can_migrate_default=True,
            can_migrate_shard=False
        )

    def test_specific_database_only_on_datbase(self):
        self.assert_allow_migrate(
            app_label='tests',
            model_name='ShardedTestModelIDs',
            can_migrate_default=False,
            can_migrate_shard=False,
            migratable_db='app_shard_001'
        )

    def test_specific_sharded_model_only_on_shards(self):
        self.assert_allow_migrate(
            app_label='tests',
            model_name='TestModel',
            can_migrate_default=False,
            can_migrate_shard=True,
        )

    def test_lookup_fallback_if_migration_directory_not_the_same_as_the_model(self):
        self.assert_allow_migrate(
            app_label='tests',
            model_name='auth.User',
            can_migrate_default=True,
            can_migrate_shard=False,
        )


class RouterForPostgresIDFieldTest(TransactionTestCase):

    def setUp(self):
        self.sut = ShardedRouter()
        self.user = PostgresShardUser.objects.create_user(username='username', password='pwassword', email='test@example.com')

    @unittest.skipIf(settings.DATABASES['default']['ENGINE'] not in Backends.POSTGRES, "Not a postgres backend")
    def test_postgres_sharded_id_can_be_queried_without_using_and_without_sharded_by(self):
        created_model = PostgresCustomIDModel.objects.create(random_string='Test String', user_pk=self.user.id)
        self.assertTrue(getattr(created_model, 'id'))

        self.assertTrue(isinstance(PostgresCustomIDModel._meta.pk, PostgresShardGeneratedIDField))

        self.assertTrue(isinstance(PostgresCustomIDModel.objects, ShardManager))

        instance = PostgresCustomIDModel.objects.get(id=created_model.id)
        self.assertEqual(created_model._state.db, instance._state.db)

        instance = PostgresCustomIDModel.objects.get(pk=created_model.id)
        self.assertEqual(created_model._state.db, instance._state.db)

    @unittest.skipIf(settings.DATABASES['default']['ENGINE'] not in Backends.POSTGRES, "Not a postgres backend")
    def test_shard_extracted_correctly(self):
        created_model = PostgresCustomIDModel.objects.create(random_string='Test String', user_pk=self.user.pk)
        self.assertEqual(self.user.shard, self.sut.get_shard_for_postgres_pk_field(PostgresCustomIDModel, created_model.id))



from django.conf import settings
from django.test import TestCase
from django.utils.six.moves import xrange

from django_sharding_library.sharding_functions import (
    BaseBucketingStrategy,
    BaseShardedModelBucketingStrategy,
    RandomBucketingStrategy,
    RoundRobinBucketingStrategy,
    ModBucketingStrategy,
    SavedModBucketingStrategy
)


class BaseBucketingStrategyTestCase(TestCase):

    def test_get_shards(self):
        sut = BaseBucketingStrategy(shard_group='default')
        result = sut.get_shards(settings.DATABASES)
        expected_result = ['app_shard_001', 'app_shard_002']

        self.assertEqual(sorted(result), expected_result)

    def test_pick_shard_method_defined_but_unimplimented(self):
        from django.contrib.auth import get_user_model
        User = get_user_model()
        sut = BaseBucketingStrategy(shard_group='default')
        with self.assertRaises(NotImplementedError):
            sut.pick_shard(User)

    def test_get_shard_method_defined_but_unimplimented(self):
        from django.contrib.auth import get_user_model
        User = get_user_model()
        sut = BaseBucketingStrategy(shard_group='default')
        with self.assertRaises(NotImplementedError):
            sut.get_shard(User)


class BaseShardedModelBucketingStrategyTestCase(TestCase):

    def test_get_shard_reads_shard_from_model(self):
        class FakeModel(object):
            django_sharding__shard_field = 'whatever_field'
            whatever_field = 'cool_guy_shard'

        sut = BaseShardedModelBucketingStrategy(shard_group='deault')
        self.assertEqual(sut.get_shard(FakeModel()), 'cool_guy_shard')


class RoundRobinBucketingStrategyTestCase(TestCase):

    def test_picking_is_cyclic(self):
        class FakeUser():
            username = None

        expected_cycle = ['app_shard_001', 'app_shard_002']

        sut = RoundRobinBucketingStrategy(shard_group='default', databases=settings.DATABASES)
        resulting_shards = [sut.pick_shard(FakeUser()) for i in xrange(100)]

        self.assertEqual(len(set([resulting_shards[i] for i in xrange(0, 100, 2)])), 1)
        self.assertEqual(len(set([resulting_shards[i] for i in xrange(1, 100, 2)])), 1)

        resulting_cycled_shard = resulting_shards[:2]
        resulting_cycled_shard.sort()

        self.assertEqual(resulting_cycled_shard, expected_cycle)

    def test_shard_saved_to_user_when_used_with_model_and_signal(self):
        from django.contrib.auth import get_user_model
        user = get_user_model().objects.create_user(
            username='username', password='pwassword', email='test@example.com'
        )
        self.assertIsNotNone(user.shard)

    def test_shard_retrieve_shard_from_user_when_used_with_model_and_signal(self):
        from django.contrib.auth import get_user_model
        user = get_user_model().objects.create_user(
            username='username', password='pwassword', email='test@example.com'
        )
        sut = RoundRobinBucketingStrategy(shard_group='default', databases=settings.DATABASES)

        self.assertEqual(user.shard, sut.get_shard(user))
        self.assertIsNotNone(user.shard)

    def test_user_round_robin_when_used_with_model_and_signal(self):
        from django.contrib.auth import get_user_model
        shards = []
        for i in xrange(4):
            username = 'username{}'.format(i)
            email = 'username{}@example.com'.format(i)
            shards.append(
                get_user_model().objects.create_user(
                    username=username, password='pwassword', email=email
                ).shard
            )

        self.assertIsNotNone(shards[0])
        self.assertIsNotNone(shards[1])
        self.assertEqual(shards[0], shards[2])
        self.assertEqual(shards[1], shards[3])


class RandomBucketingStrategyTestCase(TestCase):

    def test_pick_shard_picks_valid_shards(self):
        sut = RandomBucketingStrategy(shard_group='default', databases=settings.DATABASES)

        for i in xrange(100):
            self.assertEqual(
                settings.DATABASES[sut.pick_shard(1)]['SHARD_GROUP'],
                'default'
            )


class ModBucketingStrategyTestCase(TestCase):

    def test_pick_shard_and_get_shard_equals_hash_shard_using_mod_of_pk(self):
        sut = ModBucketingStrategy(shard_group='default', databases=settings.DATABASES)
        self.assertEqual(len(sut.shards), 2)

        class FakeModel(object):
            pk = 0

        model = FakeModel()
        for i in xrange(100):
            model.pk = i
            expected_shard = sut.shards[hash(str(i)) % 2]
            self.assertEqual(sut.get_shard(model), expected_shard)
            self.assertEqual(sut.pick_shard(model), expected_shard)


class SavedModBucketingStrategyTestCase(TestCase):

    def test_pick_shard_equals_hash_shard_using_mod_of_pk_and_get_reads_from_model(self):
        sut = SavedModBucketingStrategy(shard_group='default', databases=settings.DATABASES)
        self.assertEqual(len(sut.shards), 2)

        class FakeModel(object):
            pk = 0
            django_sharding__shard_field = 'whatever_field'
            whatever_field = 'cool_guy_shard'

        model = FakeModel()
        for i in xrange(100):
            model.pk = i
            expected_shard = sut.shards[hash(str(i)) % 2]
            self.assertEqual(sut.pick_shard(model), expected_shard)
            self.assertEqual(sut.get_shard(model), 'cool_guy_shard')



from django.contrib.auth.models import AbstractUser
from django.db import models
from django.conf import settings
from django_sharding_library.decorators import model_config
from django_sharding_library.fields import TableShardedIDField, ShardForeignKeyStorageField, PostgresShardGeneratedIDField
from django_sharding_library.models import ShardedByMixin, ShardStorageModel, TableStrategyModel
from django_sharding_library.constants import Backends


# A model for use with a sharded model to generate pk's using
# an autoincrement field on the backing TableStrategyModel.
# This one is initialized but used in no strategy.
class ShardedModelIDs(TableStrategyModel):
    pass


# An implimentation of the extension of a the Django user to add
# the mixin provided in order to save the shard on the user.
class User(AbstractUser, ShardedByMixin):
    django_sharding__shard_group = 'default'


# An implimentation of the extension of a the Django user to add
# the mixin provided in order to save the shard on the user.
class PostgresShardUser(AbstractUser, ShardedByMixin):
    shard_group = 'postgres'
    django_sharding__shard_group = 'postgres'


# A model for use with a sharded model to generate pk's using
# an autoincrement field on the backing TableStrategyModel.
# This one is initialized for use with TestModel and is stored
# on `app_shard_001`.


@model_config(database='app_shard_001')
class ShardedTestModelIDs(TableStrategyModel):
    pass


# An example of a sharded model which uses the `TableStrategy` to
# generate uuid's for its instances.


@model_config(shard_group='default', sharded_by_field='user_pk')
class TestModel(models.Model):
    id = TableShardedIDField(primary_key=True, source_table=ShardedTestModelIDs)
    random_string = models.CharField(max_length=120)
    user_pk = models.PositiveIntegerField()

    def get_shard(self):
        from django.contrib.auth import get_user_model
        return get_user_model().objects.get(pk=self.user_pk).shard

    @staticmethod
    def get_shard_from_id(user_pk):
        from django.contrib.auth import get_user_model
        user = get_user_model()
        return user.objects.get(pk=user_pk).shard


@model_config(database='default')
class UnshardedTestModel(models.Model):
    id = TableShardedIDField(primary_key=True, source_table=ShardedTestModelIDs)
    random_string = models.CharField(max_length=120)
    user_pk = models.PositiveIntegerField()

    def get_shard(self):
        from django.contrib.auth import get_user_model
        return get_user_model().objects.get(pk=self.user_pk).shard


class ShardStorageTable(ShardStorageModel):
    pass


class ShardedByForiegnKeyModel(models.Model):
    shard = ShardForeignKeyStorageField(ShardStorageTable, shard_group='default')
    random_string = models.CharField(max_length=120)
    test = models.ForeignKey(UnshardedTestModel)

    def get_shard_key(self):
        return self.test.user_pk


@model_config(database='app_shard_001')
class PostgresCustomIDModelBackupField(TableStrategyModel):
    pass


@model_config(shard_group="postgres", sharded_by_field="user_pk")
class PostgresCustomIDModel(models.Model):
    if settings.DATABASES['default']['ENGINE'] in Backends.POSTGRES:
        id = PostgresShardGeneratedIDField(primary_key=True)
    else:
        id = TableShardedIDField(primary_key=True, source_table=PostgresCustomIDModelBackupField)
    random_string = models.CharField(max_length=120)
    user_pk = models.PositiveIntegerField()

    def get_shard(self):
        return PostgresShardUser.objects.get(pk=self.user_pk).shard

    @staticmethod
    def get_shard_from_id(user_pk):
        return PostgresShardUser.objects.get(pk=user_pk).shard
